"""This module contains core functionality for working with contextual bandit environments.

This module contains the abstract interfaces for common types of contextual bandit environments,
several concrete implementations of these environments for use in experiments, and various filters that
can be applied to environments to modify them in useful ways (e.g., shuffling, scaling, and imputing).
"""

from coba.environments.core       import Environments
from coba.environments.primitives import Interaction, Environment, SafeEnvironment, EnvironmentFilter
from coba.environments.primitives import SimulatedInteraction, LoggedInteraction, GroundedInteraction
from coba.environments.filters    import Shuffle, Take, Identity, Reservoir, Riffle, Cache
from coba.environments.filters    import Sort, Scale, Cycle, Impute, Flatten, Params
from coba.environments.filters    import Binary, Sparse, Where, Noise, Grounded
from coba.environments.filters    import Repr, Finalize, Batch, BatchSafe, Chunk, Logged

from coba.environments.synthetics import LambdaSimulation, LinearSyntheticSimulation, NeighborsSyntheticSimulation
from coba.environments.synthetics import KernelSyntheticSimulation, MLPSyntheticSimulation
from coba.environments.openml     import OpenmlSimulation, OpenmlSource
from coba.environments.supervised import SupervisedSimulation, CsvSource, ArffSource, LibSvmSource, ManikSource

__all__ = [
    'Interaction',
    'SimulatedInteraction',
    'LoggedInteraction',
    'GroundedInteraction',
    'Environment',
    'SafeEnvironment',
    'Environments',
    'LambdaSimulation',
    'OpenmlSimulation',
    'OpenmlSource',
    'SupervisedSimulation',
    'CsvSource',
    'ArffSource',
    'LibSvmSource',
    'ManikSource',
    'LinearSyntheticSimulation',
    'NeighborsSyntheticSimulation',
    'KernelSyntheticSimulation',
    'MLPSyntheticSimulation',
    'EnvironmentFilter',
    'Sort',
    'Scale',
    'Cycle',
    'Impute',
    'Binary',
    'Where',
    'Shuffle',
    'Take',
    'Reservoir',
    'Noise',
    'Identity',
    'Sparse',
    'Riffle',
    'Flatten',
    'Params',
    'Cache',
    'Grounded',
    'Finalize',
    'Repr',
    'Batch',
    'BatchSafe',
    'Chunk',
    'Logged'
]



"""Global class registration and creation.

This functionality both makes it possible for other packages to easily extend COBA,
and for the core functionality to specify classes creation recipes in config files.
"""

from itertools import repeat
from importlib import reload
from typing import Dict, Any, Callable, Union, Tuple, Mapping

from coba.backports.metadata import entry_points
from coba.exceptions import CobaException

def coba_registry_class(name:str) -> Callable[[type],type]:

    def registration_decorator(cls: type) -> type:
        CobaRegistry.register(name, cls)
        return cls

    return registration_decorator

class CobaRegistry_meta(type):
    """We use a meta type so that we can have a class level property."""

    def __init__(cls, *args, **kwargs):
        cls._registry: Dict[str,type] = {}
        cls._endpoints_loaded = False
    
    @property
    def registry(cls) -> Mapping[str,type]:
        
        if not cls._endpoints_loaded:
            cls._endpoints_loaded = True
            for ep in entry_points(group='coba.register'):
                reload(ep.load()) #we use reload in case the registry has been cleared at some point

        return cls._registry

class CobaRegistry(metaclass=CobaRegistry_meta):

    @classmethod
    def clear(cls) -> None:
        cls._endpoints_loaded = False
        cls._registry.clear()

    @classmethod
    def register(cls, name:str, tipe:type) -> None:
        cls._registry[name] = tipe

class JsonMakerV1:

    def __init__(self, registry:Dict[str,type]) -> None:
        self._registry = registry

    def make(self, recipe:Any) -> Any:
        name   = ""
        args   = None
        kwargs = None
        method = "singular"

        if not self._is_valid_recipe(recipe):
            raise CobaException(f"Invalid recipe {str(recipe)}")

        if isinstance(recipe, str):
            name = recipe

        if isinstance(recipe, dict):
            mutable_recipe = dict(recipe)

            method = mutable_recipe.pop("method", "singular")
            name   = mutable_recipe.pop("name"  , ""        )
            args   = mutable_recipe.pop("args"  , None      )
            kwargs = mutable_recipe.pop("kwargs", None      )

            if len(mutable_recipe) == 1:
                name, implicit_args = list(mutable_recipe.items())[0]

                if isinstance(implicit_args, dict) and not self._is_known_recipe(implicit_args):
                    kwargs = implicit_args
                else:
                    args = implicit_args

        if method == "singular":
            return self._construct_single(recipe, name, args, kwargs)
        else:
            if not isinstance(kwargs, list): kwargs = repeat(kwargs)
            if not isinstance(args  , list):   args = repeat(args)
            return [ self._construct_single(recipe, name, a, k) for a,k in zip(args, kwargs) ]

    def _is_valid_recipe(self, recipe:Any) -> bool:

        if isinstance(recipe, str):
            return True

        if isinstance(recipe, dict):
            keywords  = ["name", "args", "kwargs", "method"]
            freewords = [ key for key in recipe if key not in keywords ]

            implicit_arg       = None if len(freewords) != 1 else recipe[freewords[0]]
            implicit_is_args   = implicit_arg is not None and not isinstance(implicit_arg,dict)
            implicit_is_kwargs = implicit_arg is not None and     isinstance(implicit_arg,dict)

            no_unknown_words = len(freewords) <= 1
            contains_name    = "name" in keywords or len(freewords) == 1
            name_collision   = "name" in recipe and len(freewords) == 1
            args_collision   = "args" in recipe and implicit_is_args
            kwargs_collision = "kwargs" in recipe and implicit_is_kwargs
            no_collisions    = not any([name_collision, args_collision, kwargs_collision])

            return no_unknown_words and contains_name and no_collisions

        return False

    def _is_known_recipe(self, recipe:Any) -> bool:

        name = None

        if isinstance(recipe, str):
            name = recipe

        if isinstance(recipe, dict):
            name = recipe.get('name', None) or [key for key in recipe if key not in ["name", "args", "kwargs", "method"]][0]

        return name in self._registry

    def _construct_or_return(self, item:Any):
        return self.make(item) if self._is_known_recipe(item) else item

    def _construct_single(self, recipe, name, args, kwargs) -> Any:
        try:
            if args is not None and not isinstance(args, list): args = [args]

            if args is not None:
                args = [ self._construct_or_return(a) for a in args ]

            if kwargs is not None:
                kwargs = { k:self._construct_or_return(v) for k,v in kwargs.items() }

            if args is not None and kwargs is not None:
                return self._registry[name](*args, **kwargs)

            elif args is not None and kwargs is None:
                try:
                    return self._registry[name](*args)
                except TypeError as e:
                    return self._registry[name](args)

            elif args is None and kwargs is not None:
                return self._registry[name](**kwargs)
            else:
                return self._registry[name]()

        except KeyError:
            raise CobaException(f"Unknown recipe {str(recipe)}")

        except Exception as e:
            raise CobaException(f"Unable to create recipe {str(recipe)}")

class JsonMakerV2:

    #In Python the distinction between arg, args and kwargs is made using (), [], {}, *, and **
    #In json we only have [] and {} so there is going to be some ambiguity unless we add syntax.

    ##########Construction Syntax#############

    #I think this might be the best... It has low syntax overhead and clear semantic meaning. 
    #The biggest problem is that it can have a ton of ambiguity. To work around that we can offer
    #the wordier "*"/"**" notation at the bottom.
    
    #arg    = <str,float,list,dict,obj>
    #args   = [arg, ... ]
    #kwargs = {<str>:arg, ...}

    #no-arg          = < "class" | {"class":[]} >
    #arg             = { "class": <str|float>           }
    #args            = { "class": args                  }
    #kwargs          = { "class": kwargs                }
    #args and kwargs = { "class": [arg,...,"**",kwargs] }

    #High syntax overhead but no ambiguity. I think ambiguity will be rare, so probably not worth the overhead.
    # {"name":"ClassName", "args":[], "kwargs":{} }

    #Low syntax overhead, no ambiguity but I think the semantic meaning is less clear.
    #["ClassName",args,kwargs]

    #This syntax I am confident I don't like. I don't like that the levels are inconsistent (i.e., outer kw are inner args).
    #Plus, determining which key in the top-level dict is a classname and which is a kw seems tricky.
    #{ class:[], kw:arg, kw:arg }

    ###################Foreach Syntax#############################

    #I think I like this syntax more than the option below.
    #This gives the entire for collection up front and filling
    #in args is only a matter of string checks
    # {"class":"$"       , "for": [1,2,3,4]             }
    # {"class":"$"       , "for": {"zip":[[1,2],[3,4]]} }
    # {"class":{"kw":"$"}, "for": [1,2,3,4]             }

    #This has low syntax overhead but parsing and semantic reading much more difficult.
    # {class: {"for":[1,2,3,4]}             } (arg foreach construction)
    # {class: [{"for":[1,2]},{"for":[3,4]}] } (args foreach construction)
    # {class: {kw:{"for":[1,2]}}            } (kwargs foreach construction)

    def __init__(self, registry:Dict[str,type]) -> None:
        self._registry = registry

    def make(self, recipe:Union[dict,str], strict:bool = True) -> Any:
        if self._makeable(recipe):
            klass = recipe if isinstance(recipe,str) else list(recipe.keys()-{'for'})[0] if isinstance(recipe,dict) else None

            if isinstance(recipe,str):
                return self._registry[klass]()

            if isinstance(recipe,dict) and 'for' not in recipe:
                args,kwargs = self._construct_args(recipe[klass])
                return self._registry[klass](*args,**kwargs)

            if isinstance(recipe,dict) and 'for' in recipe:
                items = []
                for value in self.make(recipe['for'], strict=False):
                    args,kwargs = self._construct_args(self._fill_template(recipe[klass], value))
                    items.append(self._registry[klass](*args,**kwargs))
                return items 

        elif strict:
            raise CobaException(f"We were unable to make {recipe}.") # raise helpful exception

        else:
            return recipe

    def _makeable(self, recipe:Union[dict,str]) -> bool:
        item_len = 1 if isinstance(recipe,str) else len(recipe) if isinstance(recipe,dict) else 0
        item_cls = recipe if isinstance(recipe,str) else list(recipe.keys()-{'for'})[0] if isinstance(recipe,dict) else None
        item_for = 'for' in recipe.keys() if isinstance(recipe,dict) else False

        return (item_len == 1 and item_cls in self._registry) or (item_len == 2 and item_cls in self._registry and item_for)
    
    def _construct_args(self, args:Union[list,dict,str,int,float,None]) -> Tuple[list,dict]:
        if isinstance(args,dict):
            return [], {k:self.make(v,False)  for k,v in args.items()}
        elif isinstance(args,list) and "**" not in args:
            return [ self.make(a,False) for a in args], {}
        elif isinstance(args,list) and "**" == args[-2] and isinstance(args[-1],dict) :
            return self._construct_args(args[:-2])[0], self._construct_args(args[-1])[1]
        else:
            return [ self.make(args,False) ], {}
    
    def _fill_template(self, temp:Union[list,dict,str,int,float,None], value:Union[list,dict,str,int,float,None]) -> Union[list,dict,str,int,float,None]:
        if isinstance(temp,dict):
            return { self._fill_template(k,value):self._fill_template(v,value) for k,v in temp.items() }
        if isinstance(temp,list):
            return [ self._fill_template(i, value) for i in temp ]
        if isinstance(temp,str) and temp.startswith("$"):
            return value if temp=="$" else value[int(temp[1:])]
        return temp



from pathlib import Path
from itertools import product
from typing import Sequence, Optional, Union, overload, Tuple

from coba.pipes import Pipes
from coba.learners import Learner
from coba.environments import Environment
from coba.multiprocessing import CobaMultiprocessor
from coba.contexts import CobaContext, ExceptLog, StampLog, NameLog, DecoratedLogger, NullLogger
from coba.exceptions import CobaException

from coba.experiments.process import CreateWorkItems,  RemoveFinished, ChunkByChunk, MaxChunkSize, ProcessWorkItems
from coba.experiments.tasks   import EnvironmentTask, EvaluationTask, LearnerTask
from coba.experiments.tasks   import SimpleLearnerInfo, SimpleEnvironmentInfo, SimpleEvaluation
from coba.experiments.results import Result, TransactionIO

class Experiment:
    """An Experiment using a collection of environments and learners."""

    @overload
    def __init__(self,
        environments    : Union[Environment, Sequence[Environment]],
        learners        : Union[Learner,Sequence[Learner]],
        *,
        description     : str             = None,
        learner_task    : LearnerTask     = SimpleLearnerInfo(),
        environment_task: EnvironmentTask = SimpleEnvironmentInfo(),
        evaluation_task : EvaluationTask  = SimpleEvaluation()) -> None:
        """Instantiate an Experiment.

        Args:
            environments: The collection of environments to use in the experiment.
            learners: The collection of learners to use in the experiment.
            description: A description of the experiment for documentaiton purposes.
            learner_task: A task which describes a learner.
            environment_task: A task which describes an environment.
            evaluation_task: A task which evaluates a learner on an environment.
        """

    @overload
    def __init__(self,
        eval_pairs      : Sequence[Tuple[Learner,Environment]],
        *,
        description     : str             = None,
        learner_task    : LearnerTask     = SimpleLearnerInfo(),
        environment_task: EnvironmentTask = SimpleEnvironmentInfo(),
        evaluation_task : EvaluationTask  = SimpleEvaluation()) -> None:
        ...
        """Instantiate an Experiment.

        Args:
            eval_pairs: The collection of learners with their evaluation environments.
            learner_task: A task which describes a learner.
            environment_task: A task which describes an environment.
            evaluation_task: A task which evaluates a learner on an environment.
        """

    def __init__(self, *args, **kwargs) -> None:
        """Instantiate an Experiment."""

        if len(args) == 2:
            envs = [args[0]] if hasattr(args[0],'read') else args[0]
            lrns = [args[1]] if hasattr(args[1],'predict') else args[1]
            args = [list(zip(*reversed(list(zip(*product(envs,lrns))))))]

        self._pairs            = args[0]
        self._description      = kwargs.get('description',None)
        self._learner_task     = kwargs.get('learner_task', SimpleLearnerInfo())
        self._environment_task = kwargs.get('environment_task', SimpleEnvironmentInfo())
        self._evaluation_task  = kwargs.get('evaluation_task', SimpleEvaluation())

        if any([lrn is None for lrn,_ in self._pairs]):
            raise CobaException("A Learner was given whose value was None, which can't be processed.")

        if any([env is None for _,env in self._pairs]):
            raise CobaException("An Environment was given whose value was None, which can't be processed.")

        self._processes        : Optional[int] = None
        self._maxchunksperchild: Optional[int] = None
        self._maxtasksperchunk : Optional[int] = None

    def config(self,
        processes: int = None,
        maxchunksperchild: Optional[int] = None,
        maxtasksperchunk: Optional[int] = None) -> 'Experiment':
        """Configure how the experiment will be executed.

        A value of `None` for any item means the CobaContext.experiment will be used.

        Args:
            processes: The number of processes to create for evaluating the experiment.
            maxchunksperchild: The number of chunks each process evaluate before being restarted. A 
                value of 0 means that all processes will survive until the end of the experiment.
            maxtasksperchunk: The maximum number of tasks a chunk can have. If a chunk has too many 
                tasks it will be split into smaller chunks. A value of 0 means that chunks are never
                broken down into smaller chunks.
        """

        assert processes is None or processes > 0, "The given number of processes is invalid. Must be greater than 0."
        assert maxchunksperchild is None or maxchunksperchild >= 0, "The given number of chunks per child is invalid. Must be greater than or equal to 0 (0 for infinite)."
        assert maxtasksperchunk is None or maxtasksperchunk >= 0, "The given number of tasks per chunk is invalid. Must be greater than or equal to 0 (0 for infinite)."

        self._processes         = processes
        self._maxchunksperchild = maxchunksperchild
        self._maxtasksperchunk  = maxtasksperchunk

        return self

    @property
    def processes(self) -> int:
        """The number of processes to use when evaluating the experiment."""
        return self._processes if self._processes is not None else CobaContext.experiment.processes

    @property
    def maxchunksperchild(self) -> int:
        """The number of tasks chunks to perform per process before restarting an evaluation process."""
        return self._maxchunksperchild if self._maxchunksperchild is not None else CobaContext.experiment.maxchunksperchild

    @property
    def maxtasksperchunk(self) -> int:
        """The maximum number of tasks allowed in a chunk before breaking a chunk into smaller chunks."""
        return self._maxtasksperchunk if self._maxtasksperchunk is not None else CobaContext.experiment.maxtasksperchunk

    def run(self, result_file:str = None, quiet:bool = False) -> Result:
        """Run the experiment and return the results.

        Args:
            result_file: The file for writing and restoring results .
        """
        mp, mc, mt = self.processes, self.maxchunksperchild, self.maxtasksperchunk

        if quiet: 
            old_logger = CobaContext.logger
            CobaContext.logger = NullLogger()

        if mp > 1 or mc != 0:
            #Add name so that we know which process-id the logs came from in addition to the time of the log
            CobaContext.logger = DecoratedLogger([ExceptLog()], CobaContext.logger, [NameLog(), StampLog()])
        else:
            CobaContext.logger = DecoratedLogger([ExceptLog()], CobaContext.logger, [StampLog()])

        if result_file and Path(result_file).exists():
            CobaContext.logger.log("Restoring existing experiment logs...")
            restored = Result.from_file(result_file)
        else:
            restored = None

        n_given_learners     = len(set([l for l,_ in self._pairs]))
        n_given_environments = len(set([e for _,e in self._pairs]))

        if restored:
            assert n_given_learners     == restored.experiment.get('n_learners',n_given_learners)        , "The current experiment doesn't match the given transaction log."
            assert n_given_environments == restored.experiment.get('n_environments',n_given_environments), "The current experiment doesn't match the given transaction log."

        workitems  = CreateWorkItems(self._pairs, self._learner_task, self._environment_task, self._evaluation_task)
        unfinished = RemoveFinished(restored)
        chunk      = ChunkByChunk()
        max_chunk  = MaxChunkSize(mt)
        sink       = TransactionIO(result_file)

        single_process = ProcessWorkItems()
        multi_process  = Pipes.join(chunk, max_chunk, CobaMultiprocessor(ProcessWorkItems(), mp, mc))
        process        = multi_process if mp > 1 or mc != 0 else single_process

        try:
            if not restored: sink.write([["T0", {'n_learners':n_given_learners, 'n_environments':n_given_environments, 'description':self._description }]])
            Pipes.join(workitems, unfinished, process, sink).run()

        except KeyboardInterrupt: # pragma: no cover
            CobaContext.logger.log("Experiment execution was manually aborted via Ctrl-C")

        except Exception as ex: # pragma: no cover
            CobaContext.logger.log(ex)

        if isinstance(CobaContext.logger, DecoratedLogger):
            CobaContext.logger = CobaContext.logger.undecorate()

        if quiet: CobaContext.logger = old_logger

        return sink.read()

    def evaluate(self, result_file:str = None) -> Result:
        """Evaluate the experiment and return the results (this is a backwards compatible proxy for the run method).

        Args:
            result_file: The file for writing and restoring results .
        """

        return self.run(result_file=result_file)



from abc import ABC, abstractmethod
from typing import Any, TypeVar, Generic, Mapping

_T_out = TypeVar("_T_out", bound=Any, covariant    =True)
_T_in  = TypeVar("_T_in" , bound=Any, contravariant=True)

class Pipe:

    @property
    def params(self) -> Mapping[str,Any]:
        """Parameters describing the pipe."""
        return { }

    def __str__(self) -> str:
        return str(self.params)

class Source(ABC, Pipe, Generic[_T_out]):
    """A pipe that can be read."""

    @abstractmethod
    def read(self) -> _T_out:
        """Read the item."""
        ...

class Filter(ABC, Pipe, Generic[_T_in, _T_out]):
    """A pipe that can modify an item."""

    @abstractmethod
    def filter(self, item: _T_in) -> _T_out:
        """Filter the item."""
        ...

class Sink(ABC, Pipe, Generic[_T_in]):
    """A pipe that writes item."""

    @abstractmethod
    def write(self, item: _T_in) -> None:
        """Write the item."""
        ...



from typing import Union, Sequence, Mapping

Context = Union[None, str, int, float, Sequence, Mapping]
Action  = Union[str, int, float, Sequence, Mapping]
Actions = Sequence[Action]
AIndex  = int

class Batch(list):
    pass



from abc import ABC, abstractmethod
from collections import abc
from typing import Union, Sequence, Any, Iterator

from coba.primitives.semantic import Action, AIndex, Batch

class Feedback(ABC):
    @abstractmethod
    def eval(self, arg: Union[Action,AIndex]) -> Any:
        ...

class SequenceFeedback(Feedback):
    def __init__(self, values: Sequence[Any]) -> None:
        self._values = values

    def eval(self, arg: AIndex) -> Any:
        return self._values[arg]

    def __getitem__(self, index: int) -> Any:
        return self._values[index]
    
    def __len__(self) -> int:
        return len(self._values)

    def __eq__(self, o: object) -> bool:
        return isinstance(o,abc.Sequence) and list(o) == list(self._values)

class BatchFeedback(Batch):
    def eval(self, actions: Sequence[Action]) -> Sequence[Any]:
        return list(map(lambda f,a: f.eval(a), self, actions))



import re
import collections
import collections.abc

from copy import copy
from pathlib import Path
from numbers import Number
from operator import truediv, sub, gt, itemgetter
from abc import abstractmethod
from itertools import chain, repeat, accumulate, groupby, count
from typing import Any, Dict, List, Set, Tuple, Optional, Sequence, Hashable, Iterable, Iterator, Union, Type, Callable, NamedTuple
from coba.backports import Literal

from coba.environments import Environment
from coba.statistics import Mean, StandardDeviation, StandardErrorOfMean, BootstrapConfidenceInterval, BinomialConfidenceInterval, PointAndInterval
from coba.contexts import CobaContext
from coba.exceptions import CobaException
from coba.utilities import PackageChecker
from coba.pipes import Pipes, Sink, Source, JsonEncode, JsonDecode, DiskSource, DiskSink, IterableSource, ListSink, Foreach

def exponential_moving_average(values:Sequence[float], span:int=None) -> Iterable[float]:
    #exponential moving average identical to Pandas df.ewm(span=span).mean()
    alpha = 2/(1+span)
    cumwindow  = list(accumulate(values          , lambda a,v: v + (1-alpha)*a))
    cumdivisor = list(accumulate([1.]*len(values), lambda a,v: v + (1-alpha)*a)) #type: ignore
    return map(truediv, cumwindow, cumdivisor)

def moving_average(values:Sequence[float], span:int=None) -> Iterable[float]:

    if span == 1: 
        return values

    if span is None or span >= len(values):
        return (a/n for n,a in enumerate(accumulate(values),1))

    window_sums  = accumulate(map(sub, values, chain(repeat(0,span),values)))
    window_sizes = chain(range(1,span), repeat(span))

    return map(truediv,window_sums,window_sizes)

class Table:
    """A container class for storing tabular data."""

    def __init__(self, name:str, primary_cols: Sequence[str], rows: Sequence[Dict[str,Any]], preferred_cols: Sequence[str] = []):
        """Instantiate a Table.

        Args:
            name: The name of the table. Used for display purposes.
            primary_cols: Table columns used to make each row's tuple "key".
            rows: The actual rows that should be stored in the table. Each row is required to contain the given primary_cols.
            preferred_cols: A list of columns that we prefer be displayed immediately after primary columns. All remaining
                columns (i.e., neither primary nor preferred) will be ordered alphabetically.
        """
        self._name    = name
        self._primary = primary_cols

        for row in rows:
            assert len(row.keys() & primary_cols) == len(primary_cols), 'A Table row was provided without a primary key.'

        all_columns: Set[str] = set()
        for row in rows:
            all_columns |= {'index'} if '_packed' in row else set()
            all_columns |= row.keys()-{'_packed'}
            all_columns |= all_columns.union(row.get('_packed',{}).keys())

        col_priority = list(chain(primary_cols + ['index'] + preferred_cols + sorted(all_columns)))

        self._columns = sorted(all_columns, key=col_priority.index)
        self._rows_keys: Dict[Hashable, None         ] = {}
        self._rows_flat: Dict[Hashable, Dict[str,Any]] = {}
        self._rows_pack: Dict[Hashable, Dict[str,Any]] = {}

        for row in rows:
            row_key  = row[primary_cols[0]] if len(primary_cols) == 1 else tuple(row[col] for col in primary_cols)
            row_pack = row.pop('_packed',{})
            row_flat = row

            if row_pack:
                row_pack['index'] = list(range(1,len(list(row_pack.values())[0])+1))

            self._rows_keys[row_key] = None
            self._rows_pack[row_key] = row_pack
            self._rows_flat[row_key] = row_flat

        self._rows_keys = collections.OrderedDict(zip(sorted(list(self._rows_keys.keys())),repeat(None)))

    @property
    def name(self) -> str:
        """The name of the table."""
        return self._name

    @property
    def keys(self) -> Sequence[Hashable]:
        """Keys for accessing data in the table."""
        return list(self._rows_keys.keys())

    @property
    def columns(self) -> Sequence[str]:
        """The columns in the table."""
        return self._columns

    @property
    def dtypes(self) -> Sequence[Type[Union[int,float,bool,object]]]:
        """The dtypes for the columns in the table."""
        flats = self._rows_flat
        packs = self._rows_pack

        columns_packed = [ any([ col in packs[key] for key in self.keys]) for col in self.columns ]
        columns_values = [ [flats[key].get(col, packs[key].get(col, self._default(col))) for key in self.keys] for col in self.columns ]

        return [ self._infer_type(column_packed, column_values) for column_packed, column_values in zip(columns_packed,columns_values)]

    def filter(self, row_pred:Callable[[Dict[str,Any]],bool] = None, **kwargs) -> 'Table':
        """Filter to specific rows.

        Args:
            pred: A predicate that returns true for row dictionaries that should be kept.
            kwargs: key value pairs where the key is the column and the value indicates what
                value a row should have in that column to be kept. Keeping logic depends on
                the row value type and the kwargs value type. If kwarg value == row value keep
                the row. If kwarg value is callable pass the row value to the predicate. If
                kwarg value is a collection keep the row if the row value is in the collection.
                If kwarg value is a string apply a regular expression match to the row value.
        """

        def satisifies_filter(col_filter,col_value):
            if col_filter == col_value:
                return True

            if isinstance(col_filter,Number) and isinstance(col_value,str):
                return re.search(f'(\D|^){col_filter}(\D|$)', col_value)

            if isinstance(col_filter,str) and isinstance(col_value,str):
                return re.search(col_filter, col_value)

            if callable(col_filter):
                return col_filter(col_value)

            return False

        def satisfies_all_filters(key):
            row = self[key]

            row_filter_results = [ row_pred is None or row_pred(row) ]
            col_filter_results = [ ]

            for col,col_filter in kwargs.items():

                if isinstance(col_filter,collections.abc.Container) and not isinstance(col_filter,str):
                    col_filter_results.append(row[col] in col_filter or any([satisifies_filter(cf,row[col]) for cf in col_filter]))
                else:
                    col_filter_results.append(satisifies_filter(col_filter,row.get(col,self._default(col)) ))

            return all(row_filter_results+col_filter_results)

        new_result = copy(self)
        new_result._rows_keys = collections.OrderedDict(zip(filter(satisfies_all_filters,self.keys), repeat(None)))

        return new_result

    def to_pandas(self) -> Any:
        """Turn the Table into a Pandas data frame."""

        PackageChecker.pandas("Table.to_pandas")
        import pandas as pd #type: ignore
        import numpy as np  #type: ignore #pandas installs numpy so if we have pandas we have numpy

        col_numpy = { col: np.empty(len(self), dtype=dtype) for col,dtype in zip(self.columns,self.dtypes)}

        row_index = 0

        for key in self.keys:

            flat = self._rows_flat[key]
            pack = self._rows_pack[key]

            pack_size = 1 if not pack else len(pack['index'])

            for col in self.columns:
                if col in pack:
                    val = pack[col]

                elif col in flat:
                    if isinstance(flat[col], (tuple,list)):
                        val = [flat[col]]
                    else:
                        val = flat[col]

                else:
                    val = self._default(col)

                col_numpy[col][row_index:(row_index+pack_size)] = val

            row_index += pack_size

        return pd.DataFrame(col_numpy, columns=self.columns)

    def to_dicts(self) -> Sequence[Dict[str,Any]]:
        """Turn the Table into a sequence of tuples."""

        dicts = []

        for key in self.keys:

            flat = self._rows_flat[key]
            pack = self._rows_pack[key]

            if not pack:
                dicts.append( { col:flat.get(col,self._default(col)) for col in self.columns } )
            else:
                tuples = list(zip(*[pack.get(col,repeat(flat.get(col,self._default(col)))) for col in self.columns]))
                dicts.extend([ dict(zip(self.columns,t)) for t in tuples ])

        return dicts

    def _default(self, column:str) -> Any:
        return [1] if column == "index" else None

    def _infer_type(self, is_packed: bool, values: Sequence[Any]) -> Type[Union[int,float,bool,object]]:

        types: List[Optional[Type[Any]]] = []

        to_type = lambda value: None if value is None else type(value)

        for value in values:
            if is_packed and isinstance(value, (list,tuple)):
                types.extend([to_type(v) for v in value])
            else:
                types.append(to_type(value))

        return self._resolve_types(types)

    def _resolve_types(self, types: Sequence[Optional[Type[Any]]]) -> Type[Union[int,float,bool,object]]:
        types = list(set(types))

        if len(types) == 1 and types[0] in [dict,str]:
            return object

        if len(types) == 1 and types[0] in [int,float,bool]:
            return types[0]

        if all(t in [None,int,float] for t in types):
            return float

        return object

    def __iter__(self) -> Iterator[Dict[str,Any]]:
        for key in self.keys:
            yield self[key]

    def __contains__(self, key: Union[Hashable, Sequence[Hashable]]) -> bool:
        return key in self.keys

    def __str__(self) -> str:
        return str({"Table": self.name, "Columns": self.columns, "Rows": len(self)})

    def _ipython_display_(self):
        #pretty print in jupyter notebook (https://ipython.readthedocs.io/en/stable/config/integrating.html)
        print(str(self))

    def __len__(self) -> int:
        return sum([ len(self._rows_pack[key].get('index',[None])) for key in self.keys ])

    def __getitem__(self, key: Union[Hashable, Sequence[Hashable]]) -> Dict[str,Any]:
        if key not in self._rows_keys: raise KeyError(key)
        return dict(**self._rows_flat[key], **self._rows_pack[key])

class InteractionsTable(Table):
    pass

class TransactionIO_V3(Source['Result'], Sink[Any]):

    def __init__(self, log_file: Optional[str] = None, minify:bool = True) -> None:

        self._log_file = log_file
        self._minify   = minify
        self._source   = DiskSource(log_file) if log_file else IterableSource()
        self._sink     = DiskSink(log_file) if log_file else ListSink(self._source.iterable)

    def write(self, item: Any) -> None:
        if isinstance(self._sink, ListSink):
            self._sink.write(self._encode(item))
        else:
            if not Path(self._sink._filename).exists():self._sink.write('["version",3]')
            self._sink.write(JsonEncode(self._minify).filter(self._encode(item)))

    def read(self) -> 'Result':
        n_lrns   = None
        n_envs   = None
        lrn_rows = {}
        sim_rows = {}
        int_rows = {}

        if isinstance(self._source, IterableSource):
            decoded_source = self._source
        else:
            decoded_source = Pipes.join(self._source, Foreach(JsonDecode()))

        for trx in decoded_source.read():

            if not trx: continue

            if trx[0] == "benchmark":
                n_lrns = trx[1]["n_learners"]
                n_envs = trx[1]["n_simulations"]

            if trx[0] == "S":
                sim_rows[trx[1]] = trx[2]

            if trx[0] == "L":
                lrn_rows[trx[1]] = trx[2]

            if trx[0] == "I":
                int_rows[tuple(trx[1])] = trx[2]

        return Result(sim_rows, lrn_rows, int_rows, {"n_learners":n_lrns,"n_environments":n_envs})

    def _encode(self,item):
        if item[0] == "T0":
            return ['benchmark', {"n_learners":item[1], "n_simulations":item[2]}]

        if item[0] == "T1":
            return ["L", item[1], item[2]]

        if item[0] == "T2":
            return ["S", item[1], item[2]]

        if item[0] == "T3":
            rows_T = collections.defaultdict(list)

            for row in item[2]:
                for col,val in row.items():
                    if col == "rewards" : col="reward"
                    if col == "reveals" : col="reveal"
                    rows_T[col].append(val)

            return ["I", item[1], { "_packed": rows_T }]

        return None

class TransactionIO_V4(Source['Result'], Sink[Any]):

    def __init__(self, log_file: Optional[str] = None, minify:bool=True) -> None:

        self._log_file = log_file
        self._minify   = minify
        self._source   = DiskSource(log_file) if log_file else IterableSource()
        self._sink     = DiskSink(log_file)   if log_file else ListSink(self._source.iterable)

    def write(self, item: Any) -> None:
        if isinstance(self._sink, ListSink):
            self._sink.write(self._encode(item))
        else:
            if not Path(self._sink._filename).exists():self._sink.write('["version",4]')
            self._sink.write(JsonEncode(self._minify).filter(self._encode(item)))

    def read(self) -> 'Result':

        exp_dict = {}
        lrn_rows = {}
        env_rows = {}
        int_rows = {}

        if isinstance(self._source, IterableSource):
            decoded_source = self._source
        else:
            decoded_source = Pipes.join(self._source, Foreach(JsonDecode()))

        for trx in decoded_source.read():

            if not trx: continue

            if trx[0] == "experiment":
                exp_dict = trx[1]

            if trx[0] == "E":
                env_rows[trx[1]] = trx[2]

            if trx[0] == "L":
                lrn_rows[trx[1]] = trx[2]

            if trx[0] == "I":
                int_rows[tuple(trx[1])] = trx[2]

        return Result(env_rows, lrn_rows, int_rows, exp_dict)

    def _encode(self,item):
        if item[0] == "T0":
            return ['experiment', {"n_learners":item[1], "n_environments":item[2], "description":None} if len(item)==3 else item[1] ]

        if item[0] == "T1":
            return ["L", item[1], item[2]]

        if item[0] == "T2":
            return ["E", item[1], item[2]]

        if item[0] == "T3":
            rows_T = collections.defaultdict(list)

            for row in item[2]:
                for col,val in row.items():
                    if col == "rewards" : col="reward"
                    if col == "reveals" : col="reveal"
                    rows_T[col].append(val)

            return ["I", item[1], { "_packed": rows_T }]

        return None

class TransactionIO(Source['Result'], Sink[Any]):

    def __init__(self, transaction_log: Optional[str] = None) -> None:

        if not transaction_log or not Path(transaction_log).exists():
            version = None
        else:
            version = JsonDecode().filter(next(DiskSource(transaction_log).read()))[1]

        if version == 3:
            self._transactionIO = TransactionIO_V3(transaction_log)

        elif version == 4:
            self._transactionIO = TransactionIO_V4(transaction_log)

        elif version is None:
            self._transactionIO = TransactionIO_V4(transaction_log)

        else:
            raise CobaException("We were unable to determine the appropriate Transaction reader for the file.")

    def write(self, transactions: Iterable[Any]) -> None:
        for transaction in transactions:
            self._transactionIO.write(transaction)

    def read(self) -> 'Result':
        return self._transactionIO.read()

class Points(NamedTuple):
    X    : Sequence[Any]
    Y    : Sequence[float] 
    XE   : Sequence[float]  = None
    YE   : Sequence[float]  = None
    color: Union[str,int]   = None
    alpha: float            = 1
    label: Optional[str]    = None
    style: Literal['-','.'] = "-"

class Plotter:
    @abstractmethod
    def plot(self,
        ax,
        lines: Sequence[Points],
        title: str,
        xlabel: str,
        ylabel: str,
        xlim: Tuple[Optional[Number],Optional[Number]],
        ylim: Tuple[Optional[Number],Optional[Number]],
        xticks: bool,
        yticks: bool,
        out: Union[None,Literal['screen'],str]) -> None:
        pass

class MatplotlibPlotter(Plotter):

    def plot(self,
        ax,
        lines: Sequence[Points],
        title: str,
        xlabel: str,
        ylabel: str,
        xlim: Tuple[Optional[Number],Optional[Number]],
        ylim: Tuple[Optional[Number],Optional[Number]],
        xticks: bool,
        yticks: bool,
        xrotation: Optional[float],
        yrotation: Optional[float],
        out: Union[None,Literal['screen'],str]
    ) -> None:

        xlim = xlim or [None,None]
        ylim = ylim or [None,None]

        PackageChecker.matplotlib('Result.plot_learners')
        import matplotlib.pyplot as plt #type: ignore

        color_cycle = plt.rcParams['axes.prop_cycle'].by_key()['color']

        ax: plt.Axes

        bad_xlim  = xlim and xlim[0] is not None and xlim[1] is not None and xlim[0] >= xlim[1]
        bad_ylim  = ylim and ylim[0] is not None and ylim[1] is not None and ylim[0] >= ylim[1]
        bad_lines = not lines

        if bad_xlim or bad_ylim or bad_lines:
            if bad_xlim:
                CobaContext.logger.log("The xlim end is less than the xlim start. Plotting is impossible.")
            if bad_ylim:
                CobaContext.logger.log("The ylim end is less than the ylim start. Plotting is impossible.")
            if bad_lines:
                CobaContext.logger.log(f"No data was found for plotting.")
        else:

            if not ax or isinstance(ax,int):
                if 'coba' in plt.get_figlabels():
                    f = plt.figure(num="coba")
                    ax = f.add_subplot(111) if not isinstance(ax,int) else f.add_subplot(ax)
                else:
                    ax = plt.subplot(111) if not isinstance(ax,int) else plt.subplot(ax)

            in_lim = lambda v,lim: lim is None or ((lim[0] or -float('inf')) <= v and v <= (lim[1] or float('inf')))

            any_label = False
            num_coalesce = lambda x1,x2: x1 if isinstance(x1,(int,float)) else x2

            for X, Y, XE, YE, c, a, l, fmt in lines:

                if l: any_label = True

                # we remove values outside of the given lims because matplotlib won't correctly scale otherwise
                if not YE:
                    XY = [(x,y) for i,x,y in zip(count(),X,Y) if in_lim(num_coalesce(x,i),xlim) and in_lim(num_coalesce(y,i),ylim)]
                    X,Y = map(list,zip(*XY)) if XY else ([],[])
                else:
                    XYE = [(x,y,e) for i,x,y,e in zip(count(),X,Y,YE) if in_lim(num_coalesce(x,i),xlim) and in_lim(num_coalesce(y,i),ylim)]
                    X,Y,YE = map(list,zip(*XYE)) if XYE else ([],[],[])

                if isinstance(c,int): c = color_cycle[c%len(color_cycle)]

                not_err_bar = lambda E: not E or all([e is None for e in E])

                if X and Y:
                    if all(map(not_err_bar,[XE,YE])):
                        ax.plot(X, Y, fmt,  color=c, alpha=a, label=l)
                    else:
                        XE = None if not_err_bar(XE) else list(zip(*XE)) if isinstance(XE[0],tuple) else XE
                        YE = None if not_err_bar(YE) else list(zip(*YE)) if isinstance(YE[0],tuple) else YE
                        error_every = max(int(len(X)*0.05),1) if fmt == "-" else 1
                        ax.errorbar(X, Y, YE, XE, fmt, elinewidth=0.5, errorevery=error_every, color=c, alpha=a, label=l)

            if xrotation is not None:
                plt.xticks(rotation=xrotation)

            if yrotation is not None:
                plt.yticks(rotation=yrotation)

            if xlim[0] is None or xlim[1] is None:
                ax.autoscale(axis='x')

            if ylim[0] is None or ylim[1] is None:
                ax.autoscale(axis='y')

            ax.set_xlim(*xlim)
            ax.set_ylim(*ylim)

            ax.set_title(title, loc='left', pad=15)
            ax.set_ylabel(ylabel)
            ax.set_xlabel(xlabel)

            if ax.get_legend() is None: #pragma: no cover
                scale = 0.65
                box1 = ax.get_position()
                ax.set_position([box1.x0, box1.y0 + box1.height * (1-scale), box1.width, box1.height * scale])
            else:
                ax.get_legend().remove()

            if any_label:
                ax.legend(*ax.get_legend_handles_labels(), loc='upper left', bbox_to_anchor=(-.01, -.25), ncol=1, fontsize='medium') #type: ignore

            if not xticks:
                plt.xticks([])

            if not yticks:
                plt.yticks([])

            if out not in ['screen',None]:
                plt.tight_layout()
                plt.savefig(str(out), dpi=300)
                plt.show()
                plt.close()

            if out=="screen":
                plt.show()
                plt.close()

class FilterPlottingData:

    def filter(self, rows:Sequence[Dict[str,Any]], x:Sequence[str], y:str, learner_ids:Sequence[int] = None) -> Sequence[Dict[str,Any]]:

        if not rows: raise CobaException("This result doesn't contain any evaluation data to plot.")

        rows = [row for row in rows if not learner_ids or row["learner_id"] in learner_ids]

        learner_count   = len(set([row['learner_id'] for row in rows]))
        environ_counts  = collections.Counter([row['environment_id'] for row in rows])
        environ_lengths = set([len(row[y]) for row in rows])
        min_env_length  = min(environ_lengths) 

        if max(environ_counts.values()) != learner_count:
            raise CobaException("This result does not contain an environment which has been finished for every learner. Plotting has been stopped.")

        if min(environ_counts.values()) != learner_count:
            CobaContext.logger.log("This result contains environments not present for all learners. Environments not present for all learners have been excluded. To supress this warning in the future call <result>.filter_fin() before plotting.") 

        if len(environ_lengths) > 1 and x == ['index']:
            CobaContext.logger.log("This result contains environments of different lengths. The plot only includes interactions up to the shortest environment. To supress this warning in the future call <result>.filter_fin(n_interactions) before plotting.")

        complete    = lambda row: environ_counts[row['environment_id']] == learner_count
        subselect_y = lambda ys: ys[:min_env_length] if x == ['index'] else ys
        subselect_r = lambda row: {k:(row[k] if k != y else subselect_y(row[k])) for k in ['environment_id','learner_id',y]}

        return list(map(subselect_r,filter(complete,rows)))

class SmoothPlottingData:

    def filter(self, rows:Sequence[Dict[str,Any]], y:str, span:Optional[int]) -> Sequence[Dict[str,Any]]:
        e_key = "environment_id"
        l_key = "learner_id"
        return [{e_key:row[e_key], l_key:row[l_key], y:moving_average(row[y],span)} for row in rows]

class ContrastPlottingData:

    def filter(self, rows:Sequence[Dict[str,Any]], y:str, mode:str, learner_id1:int) -> Sequence[Dict[str,Any]]:

        sort_key  = lambda row: (row['environment_id'], 0 if row['learner_id']==learner_id1 else 1)
        group_key = lambda row: row['environment_id']

        if mode == "prob":
            contraster = lambda Y1,Y2: list(map(int,map(gt,Y1,Y2))) 
        elif mode == "diff": 
            contraster = lambda Y1,Y2: list(map(sub,Y1,Y2))
        else:
            contraster = lambda Y1,Y2: list(zip(Y1,Y2))

        contrast = lambda env_id, pair: {'environment_id': env_id, y: contraster(pair[0][y],pair[1][y]) }

        return [ contrast(env_id,list(pair)) for env_id, pair in groupby(sorted(rows,key=sort_key),key=group_key) ]

class TransformToXYE:

    def filter(self,
        rows: Sequence[Dict[str,Any]], 
        envs: Dict[int,Dict[str,Any]], 
        x:Sequence[str], 
        y:str,
        err:Union[str,None,PointAndInterval]) -> Sequence[Tuple[Any,float,Union[None,float,Tuple[float,float]]]]:

        if err == 'sd':
            avg = Mean()
            std = StandardDeviation()
            YE = lambda z: (avg.calculate(z), std.calculate(z))
        elif err == 'se':
            YE = StandardErrorOfMean().calculate
        elif err == "bs":
            YE = BootstrapConfidenceInterval(.95, Mean().calculate).calculate
        elif err == "bi":
            YE = BinomialConfidenceInterval('wilson').calculate
        elif isinstance(err,PointAndInterval):
            YE = err.calculate
        else:
            avg = Mean()
            YE = lambda z: (avg.calculate(z) if len(z) > 1 else z[0], None)

        iters = [ iter(row[y]) for row in rows ]
        first_val = next(iters[0])
        iters[0] = chain([first_val],iters[0])

        is_scatter = isinstance(first_val,(list,tuple))

        if x == ['index']:
            Z = zip(*iters)
            X = count(1)

            if not is_scatter:
                points = [(x,None)+YE(z) for x,z in zip(X,Z) ]
            else:
                points = [YE(z1)+YE(z2) for _,z in zip(X,Z) for z1,z2 in [tuple(zip(*z))]]
        else:
            XZ = collections.defaultdict(list)
            make_x = lambda env: env[x[0]] if len(x) == 1 else tuple(env[k] for k in x)
            for row,I in zip(rows,iters):
                XZ[str(make_x(envs[row["environment_id"]]))].append(list(I)[-1])

            if not is_scatter:
                points = [(x,None)+YE(z) for x,z in XZ.items()]
            else:
                points = [YE(z1)+YE(z2) for _,z in XZ.items() for z1,z2 in [tuple(zip(*z))]]

        return [ (x,y,xe,ye) for x,xe,y,ye in points ]

class Result:
    """A class representing the result of an Experiment."""

    @staticmethod
    def from_file(filename: str) -> 'Result':
        """Create a Result from a transaction file."""

        if not Path(filename).exists():
            raise CobaException("We were unable to find the given Result file.")

        return TransactionIO(filename).read()

    @staticmethod
    def from_logged_envs(environments: Sequence[Environment]):

        environments = [e for e in environments if e.params.get('logged')]

        envs_params = []
        lrns_params = []

        env_rows = {}
        lrn_rows = {}
        int_rows = {}

        my_mean = lambda x: sum(x)/len(x)

        for env in environments:
            
            is_batched = 'batched' in env.params
            env_params  = dict(env.params)
            lrn_params  = env_params.pop('learner')

            try:
                env_id = envs_params.index(env_params)
            except:
                env_id = len(envs_params)
                envs_params.append(env_params)
                env_rows[env_id] = env_params

            try:
                lrn_id = lrns_params.index(lrn_params)
            except:
                lrn_id = len(lrns_params)
                lrns_params.append(lrn_params)
                lrn_rows[lrn_id] = lrn_params

            if is_batched:
                results = {'_packed':{'reward': list(map(my_mean,map(itemgetter('reward'),env.read())))}}
            else:
                results = {'_packed':{'reward': list(map(itemgetter('reward'),env.read())) }}

            int_rows[(env_id,lrn_id)]= results
        
        return Result(env_rows,lrn_rows,int_rows)
        
    def __init__(self,
        env_rows: Dict[int           ,Dict[str,Any]] = {},
        lrn_rows: Dict[int           ,Dict[str,Any]] = {},
        int_rows: Dict[Tuple[int,int],Dict[str,Any]] = {},
        exp_dict: Dict[str, Any]                     = {}) -> None:
        """Instantiate a Result class.

        This constructor should never be called directly. Instead a Result file should be created
        from an Experiment and the result file should be loaded via Result.from_file(filename).
        """

        env_flat = [ { "environment_id":k,                       **v } for k,v in env_rows.items() ]
        lrn_flat = [ {                        "learner_id" :k,   **v } for k,v in lrn_rows.items() ]
        int_flat = [ { "environment_id":k[0], "learner_id":k[1], **v } for k,v in int_rows.items() ]

        self.experiment = exp_dict

        self._environments = Table            ("Environments", ['environment_id'              ], env_flat, ["source"])
        self._learners     = Table            ("Learners"    , ['learner_id'                  ], lrn_flat, ["family","shuffle","take"])
        self._interactions = InteractionsTable("Interactions", ['environment_id', 'learner_id'], int_flat, ["index","reward"])

        self._plotter = MatplotlibPlotter()

    @property
    def learners(self) -> Table:
        """The collection of learners used in the Experiment.

        The primary key of this table is learner_id.
        """
        return self._learners

    @property
    def environments(self) -> Table:
        """The collection of environments used in the Experiment.

        The primary key of this table is environment_id.
        """
        return self._environments

    @property
    def interactions(self) -> InteractionsTable:
        """The collection of interactions that learners chose actions for in the Experiment.

        The primary key of this Table is (index, environment_id, learner_id). It should be noted that the InteractionTable
        has an additional property, to_progressive_pandas() which calculates the expanding moving average or exponential
        moving average for interactions ordered by index and grouped by environment_id and learner_id.
        """
        return self._interactions

    def set_plotter(self, plotter: Plotter) -> None:
        """Manually set the underlying plotting tool. By default matplotlib is used though this can be changed."""
        self._plotter = plotter

    def copy(self) -> 'Result':
        """Create a copy of Result."""

        result = Result()

        result._environments = copy(self._environments)
        result._learners     = copy(self._learners)
        result._interactions = copy(self._interactions)

        return result

    def filter_fin(self, n_interactions: int = None) -> 'Result':
        """Filter the result to only contain data about environments with all learners and interactions.

        Args:
            n_interactions: The number of interactions at which an environment is considered complete.
        """

        def has_all(env_id):
            return all((env_id, lrn_id) in self.interactions for lrn_id in self.learners.keys)

        def has_min(env_id):
            if not n_interactions:
                return True
            else:
                return all(len(self.interactions[(env_id, lrn_id)].get('index',[])) >= n_interactions for lrn_id in self.learners.keys)

        complete_ids = [env_id for env_id in self.environments.keys if has_all(env_id) and has_min(env_id)]

        new_result               = copy(self)
        new_result._environments = self.environments.filter(environment_id=complete_ids)
        new_result._interactions = self.interactions.filter(environment_id=complete_ids)

        if n_interactions:
            new_inter = copy(new_result.interactions)
            new_inter._rows_pack = { rk: { pk: pv[0:n_interactions] for pk, pv in rv.items() } for rk,rv in new_inter._rows_pack.items() }
            new_result._interactions = new_inter

        if len(new_result.environments) == 0:
            CobaContext.logger.log(f"There was no environment which was finished for every learner.")

        return new_result

    def filter_env(self, pred:Callable[[Dict[str,Any]],bool] = None, **kwargs: Any) -> 'Result':
        """Filter the result to only contain data about specific environments.

        Args:
            pred: A predicate that returns true for learner dictionaries that should be kept.
            **kwargs: key-value pairs to filter on. To see filtering options see Table.filter.
        """

        new_result = copy(self)
        new_result._environments = new_result.environments.filter(pred, **kwargs)
        new_result._interactions = new_result.interactions.filter(lambda row: row['environment_id'] in new_result.environments)

        if len(new_result.environments) == 0:
            CobaContext.logger.log(f"No environments matched the given filter.")

        return new_result

    def filter_lrn(self, pred:Callable[[Dict[str,Any]],bool] = None, **kwargs: Any) -> 'Result':
        """Filter the result to only contain data about specific learners.

        Args:
            pred: A predicate that returns true for learner dictionaries that should be kept.
            **kwargs: key-value pairs to filter on. To see filtering options see Table.filter.
        """

        new_result = copy(self)
        new_result._learners     = new_result.learners.filter(pred, **kwargs)
        new_result._interactions = new_result.interactions.filter(learner_id=new_result.learners)

        if len(new_result.learners) == 0:
            CobaContext.logger.log(f"No learners matched the given filter.")

        return new_result

    def plot_contrast(self,
        learner_id1: int,
        learner_id2: int,
        x          : Union[str, Sequence[str]] = "environment_id",
        y          : str = "reward",
        mode       : Literal["diff","prob",'scat'] = "diff",
        span       : int = None,
        err        : Union[Literal['se','sd','bs'], None, PointAndInterval] = None,
        labels     : Sequence[str] = None,
        colors     : Sequence[str] = None,
        xlim       : Tuple[Optional[Number],Optional[Number]] = None,
        ylim       : Tuple[Optional[Number],Optional[Number]] = None,
        xticks     : bool = True,
        yticks     : bool = True,
        out        : Union[None,Literal['screen'],str] = 'screen',
        ax = None) -> None:
        """Plot a direct contrast of the performance for two learners.

        Args:
            learner_id1: The first learner to plot in the contrast.
            learner_id2: The second learner to plot in the contrast.
            x: The value to plot on the x-axis. This can either be index or environment columns to group by.
            y: The value to plot on the y-axis.
            mode: The kind of contrast plot to make: diff plots the pairwise difference, prob plots the the probability
                of learner_id1 beating learner_id2, and scatter plots learner_id1 on x-axis and learner_id2 on y axis.
            span: The number of y values to smooth together when reporting y. If this is None then the average of all y
                values up to current is shown otherwise a moving average with window size of span (the window will be
                smaller than span initially).
            err: This determines what kind of error bars to plot (if any). If `None` then no bars are plotted, if 'se'
                the standard error is shown, and if 'sd' the standard deviation is shown.
            labels: The legend labels to use in the plot. These should be in order of the actual legend labels.
            colors: The colors used to plot the learners plot.
            xlim: Define the x-axis limits to plot. If `None` the x-axis limits will be inferred.
            ylim: Define the y-axis limits to plot. If `None` the y-axis limits will be inferred.
            xticks: Whether the x-axis labels should be drawn.
            yticks: Whether the y-axis labels should be drawn.
            out: Indicate where the plot should be sent to after plotting is finished.
            ax: Provide an optional axes that the plot will be drawn to. If not provided a new figure/axes is created.
        """

        xlim = xlim or [None,None]
        ylim = ylim or [None,None]

        x = [x] if isinstance(x,str) else list(x)
        self._validate_parameters(x)

        rows = FilterPlottingData().filter(list(self.interactions), x, y, [learner_id1, learner_id2])
        rows = SmoothPlottingData().filter(rows, y, span)
        rows = ContrastPlottingData().filter(rows, y, mode, learner_id1)

        XYE = TransformToXYE().filter(rows, self.environments, x, y, err)

        if x != ['index']:
            XYE = sorted(XYE, key=lambda xye: xye[1])

        bound = .5 if mode == "prob" else 0

        win,tie,loss = [],[],[]
        for _x,_y,_xe,_ye in XYE:
            xl,xu = (0,0) if _xe is None else (_xe,_xe) if isinstance(_xe,Number) else _xe
            yl,yu = (0,0) if _ye is None else (_ye,_ye) if isinstance(_ye,Number) else _ye
            if mode != 'scat':
                (win if _y-yl > bound else loss if _y+yu < bound else tie).append((_x,_y,_xe,_ye))
            else:
                (win if _x-xl>_y+yu else loss if _y-yl>_x+xu else tie).append((_x,_y,_xe,_ye))

        colors = (colors or []) + [0,1,2]
        labels = (labels or []) + [self._full_name(learner_id1), self._full_name(learner_id2)]

        fmt = "-" if x == ['index'] else "."

        plots = []

        if loss: plots.append(Points(*zip(*loss), colors[0], 1, labels[1] + " " + f"({len(loss)})", fmt))
        if tie : plots.append(Points(*zip(*tie) , colors[1], 1, 'Tie'     + " " + f"({len(tie )})", fmt))
        if win : plots.append(Points(*zip(*win) , colors[2], 1, labels[0] + " " + f"({len(win )})", fmt))

        if mode != 'scat':
            leftmost_x  = (loss+tie+win)[ 0][0]
            rightmost_x = (loss+tie+win)[-1][0]
            plots.append(Points((leftmost_x,rightmost_x),(bound,bound), None, None , "#888", 1, None, '-'))
        else:
            m    = max([p[0] for p in (loss+tie+win)]+[p[1] for p in (loss+tie+win)]+[1])
            plots.append(Points((0,m),(0,m), None, None , "#888", 1, None, '-'))

        xrotation = 90 if x != ['index'] and len(XYE)>5 else 0
        yrotation = 0

        if mode != "scat":
            xlabel = "Interaction" if x==['index'] else x[0] if len(x) == 1 else x
            ylabel = f"{labels[0]} - {labels[1]}" if mode=="diff" else f"P({labels[0]} > {labels[1]})"
        else:
            xlabel = y
            ylabel = y

        title  = f"{ylabel} ({len(rows) if x==['index'] else len(XYE)} Environments)"

        self._plotter.plot(ax, plots, title, xlabel, ylabel, xlim, ylim, xticks, yticks, xrotation, yrotation, out)

    def plot_learners(self,
        ids   : Union[int,Sequence[int]] = None, 
        x     : Union[str,Sequence[str]] = "index",
        y     : str = "reward",
        span  : int = None,
        err   : Union[Literal['se','sd','bs'], None, PointAndInterval] = None,
        labels: Sequence[str] = None,
        colors: Union[int,Sequence[Union[str,int]]] = None,
        xlim  : Tuple[Optional[Number],Optional[Number]] = None,
        ylim  : Tuple[Optional[Number],Optional[Number]] = None,
        xticks: bool = True,
        yticks: bool = True,
        top_n: int = None,
        out   : Union[None,Literal['screen'],str] = 'screen',
        ax = None) -> None:
        """Plot the performance of multiple learners on multiple environments. It gives a sense of the expected
            performance for different learners across independent environments. This plot is valuable in gaining
            insight into how various learners perform in comparison to one another.

        Args:
            ids: Sequence of learner ids to plot (if None we will plot all learnesr in the result).
            x: The value to plot on the x-axis. This can either be index or environment columns to group by.
            y: The value to plot on the y-axis.
            span: The number of y values to smooth together when reporting y. If this is None then the average of all y
                values up to current is shown otherwise a moving average with window size of span (the window will be
                smaller than span initially).
            err: This determines what kind of error bars to plot (if any). If `None` then no bars are plotted, if 'se'
                the standard error is shown, and if 'sd' the standard deviation is shown.
            labels: The legend labels to use in the plot. These should be in order of the actual legend labels.
            colors: The colors used to plot the learners plot.
            xlim: Define the x-axis limits to plot. If `None` the x-axis limits will be inferred.
            ylim: Define the y-axis limits to plot. If `None` the y-axis limits will be inferred.
            xticks: Whether the x-axis labels should be drawn.
            yticks: Whether the y-axis labels should be drawn.
            top_n: Only plot the top_n learners. If `None` all learners will be plotted. If negative the bottom will be plotted.
            out: Indicate where the plot should be sent to after plotting is finished.
            ax: Provide an optional axes that the plot will be drawn to. If not provided a new figure/axes is created.
        """

        xlim = xlim or [None,None]
        ylim = ylim or [None,None]

        if isinstance(ids,int): ids = [ids]
        if isinstance(labels,str): labels = [labels]
        if isinstance(x,str): x = [x]

        self._validate_parameters(x)

        interactions = self.interactions if not ids else self.filter_lrn(learner_id=ids).interactions

        rows = FilterPlottingData().filter(list(interactions), x, y)
        rows = SmoothPlottingData().filter(rows, y, span)

        env_rows = self.environments
        get_key  = lambda row: row['learner_id']
        lines    = []

        style = "-" if x == ['index'] else "."

        def get_color(colors:Union[None,Sequence[Union[str,int]]], i:int):
            try:
                return colors[i] if colors else i
            except IndexError:
                return i+max(colors) if isinstance(colors[0],(int,float)) else i
            except TypeError:
                return i+colors

        def get_label(labels:Sequence[str], i:int, lrn_id:int=None):
            try:
                return labels[i] if labels else self._full_name(lrn_id)
            except:
                return self._full_name(lrn_id)

        for i, (lrn_id, lrn_rows) in enumerate(groupby(sorted(rows, key=get_key),key=get_key)):
            lrn_rows = list(lrn_rows)
            XYE      = TransformToXYE().filter(lrn_rows, env_rows, x, y, err)
            color    = get_color(colors,i)
            label    = get_label(labels,i,lrn_id)
            lines.append(Points(*zip(*XYE), color, 1, label, style))

        lines  = sorted(lines, key=lambda line: line[1][-1], reverse=True)
        labels = [l.label for l in lines]
        colors = [l.color for l in lines]
        xlabel = "Interaction" if x==['index'] else x[0] if len(x) == 1 else x
        ylabel = y.capitalize().replace("_pct"," Percent")

        title = ("Instantaneous" if span == 1 else f"Span {span}" if span else "Progressive") + f" {ylabel}"
        title = title + f" ({len(lrn_rows) if x==['index'] else len(XYE)} Environments)"

        if x != ['index']: title = f"Final {title}"

        if top_n:
            if abs(top_n) > len(lines): top_n = len(lines)*abs(top_n)/top_n
            if top_n > 0: lines = [l._replace(color=get_color(colors,i),label=get_label(labels,i)) for i,l in enumerate(lines[:top_n],0    ) ]
            if top_n < 0: lines = [l._replace(color=get_color(colors,i),label=get_label(labels,i)) for i,l in enumerate(lines[top_n:],top_n) ]

        self._plotter.plot(ax, lines, title, xlabel, ylabel, xlim, ylim, xticks, yticks, 0, 0, out)

    def _full_name(self,lrn_id:int) -> str:
        """A user-friendly name created from a learner's params for reporting purposes."""

        values = self.learners[lrn_id]
        family = values.get('family',values['learner_id'])
        params = f"({','.join(f'{k}={v}' for k,v in values.items() if k not in ['family','learner_id'])})"

        return family if params == '()' else family+params

    def __str__(self) -> str:
        return str({"Learners": len(self._learners), "Environments": len(self._environments), "Interactions": len(self._interactions) })

    def _ipython_display_(self):
        #pretty print in jupyter notebook (https://ipython.readthedocs.io/en/stable/config/integrating.html)
        print(str(self))

    def _validate_parameters(self, x:Sequence[str]):
        if 'index' in x and len(x) > 1:
            raise CobaException('The x-axis cannot contain both interaction index and environment features. Please choose one or the other.')



from typing import Sequence

class Categorical(str):
    __slots__ = ('levels','as_int','as_onehot')

    def __new__(cls, value:str, levels: Sequence[str]) -> str:
        return str.__new__(Categorical,value)

    def __init__(self, value:str, levels: Sequence[str]) -> None:
        self.levels = levels
        self.as_int = levels.index(value)
        onehot = [0]*len(levels)
        onehot[self.as_int] = 1
        self.as_onehot = tuple(onehot)

    def __repr__(self) -> str:
        return f"Categorical('{self}',{self.levels})"



"""This module contains core experiment and result functionality.

This module contains the Experiment evaluator, implementations of specific tasks that are performed
during an experiment evaluation, and the Result datastructures returned after experiment evalution.
"""

from coba.experiments.core    import Experiment
from coba.experiments.results import Result, Table
from coba.experiments.tasks   import LearnerTask, SimpleLearnerInfo
from coba.experiments.tasks   import EnvironmentTask, SimpleEnvironmentInfo, ClassEnvironmentInfo
from coba.experiments.tasks   import EvaluationTask, SimpleEvaluation

__all__ = [
    'Result',
    'Table',
    'Experiment',
    'LearnerTask',
    'EnvironmentTask',
    'EvaluationTask',
    'SimpleLearnerInfo',
    'ClassEnvironmentInfo',
    'SimpleEvaluation',
    'SimpleEnvironmentInfo',
]


from collections import abc
from math import isclose
from abc import ABC, abstractmethod
from itertools import repeat
from typing import Any, Sequence, Union, Tuple, Callable, Mapping, Optional, Type

from coba.exceptions import CobaException
from coba.random import CobaRandom
from coba.primitives import Context, Action, Actions, AIndex
from coba.primitives import Batch, Dense, Sparse, HashableDense, HashableSparse

kwargs = Mapping[str,Any]
Score  = float
PMF    = Sequence[float]
PDF    = Callable[[Union[Action,AIndex]],float]

class Probs(list):
    pass

class ActionScore(tuple):
    def __new__(self, action: Union[Action,AIndex], score: Score):
        return tuple.__new__(ActionScore, (action, score))

Prediction = Union[
    PDF,
    Probs,
    ActionScore,
    Tuple[PDF                       , kwargs],
    Tuple[Union[Action,AIndex],Score, kwargs],
    Tuple[ActionScore               , kwargs],
]

class Learner(ABC):
    """The Learner interface for contextual bandit learning."""

    @property
    def params(self) -> Mapping[str,Any]: # pragma: no cover
        """Parameters describing the learner (used for descriptive purposes only).

        Remarks:
            These will become columns in the learners table of experiment results.
        """
        return {}

    @abstractmethod
    def predict(self, context: Context, actions: Sequence[Action]) -> Prediction:
        """Predict which action to take in the context.

        Args:
            context: The current context. It will either be None (multi-armed bandit),
                a value (a single feature) a hashable tuple (dense context), or a
                hashable dictionary (sparse context).
            actions: The current set of actions to choose from in the given context.
                Each action will either be a value (a single feature), a hashable tuple
                (dense context), or a hashable dictionary (sparse context)..

        Returns:
            A Prediction. Several prediction formats are supported. See the type-hint for these.
        """
        ...

    @abstractmethod
    def learn(self,
        context: Context,
        actions: Actions,
        action: Union[Action,AIndex],
        feedback: Union[float,Any],
        score: float,
        **kwargs) -> None:
        """Learn about the action taken in the context.

        Args:
            context: The context in which the action was taken.
            actions: The set of actions chosen from.
            action: Either the action that was chosen or the index of the action that was chosen.
            feedback: This will be reward for contextual bandit problems and feedback for IGL problems.
            score: This will be the probability for the action taken if a PMF/PDF is returned by predict.
                 It will be the score if an action-score pair is returned by predict. And it will be the
                 probability if off-policy learning is being performed on LoggedInteractions.
            **kwargs: Optional information returned with the prediction.
        """
        ...

class SafeLearner(Learner):
    """A wrapper for learner-likes that guarantees interface consistency."""

    def __init__(self, learner: Learner, seed:int=1) -> None:
        """Instantiate a SafeLearner.

        Args:
            learner: The learner we wish to make sure has the expected interface
        """

        self._learner     = learner if not isinstance(learner, SafeLearner) else learner._learner
        self._rng         = CobaRandom(seed)
        self._batched     = None
        self._batched_lrn = None
        self._batched_mjr = None
        self._pred_type   = None #1==PDF,2==PMF,3==Action/Score
        self._with_info   = None
        self._learn_type  = None #3==current,2==old with info,1==old without info

    @property
    def full_name(self) -> str:
        """A user-friendly name created from a learner's params for reporting purposes."""

        params = dict(self.params)
        family = params.pop("family")

        if len(params) > 0:
            return f"{family}({','.join(f'{k}={v}' for k,v in params.items())})"
        else:
            return family

    @property
    def params(self) -> Mapping[str, Any]:
        try:
            params = self._learner.params
            params = params if isinstance(params,dict) else params()
        except AttributeError:
            params = {}

        if "family" not in params:
            params["family"] = self._learner.__class__.__name__

        return params

    def predict(self, context: Context, actions: Actions) -> Tuple[Union[AIndex,Action],Score,kwargs]:

        pred      = self._safe_predict(context,actions)
        pred_type = self._pred_type
        batched   = self._batched

        #if not batched everything just works

        #if batched and what we are given are pmfs
            #we need the prediction orderd by batch first

        #if batched and what we are given are actions cores
            #we need the prediction ordered by batch second

        if pred_type is None and not batched: # first call only
            self._determine_pred_format(pred,actions)
            pred_type = self._pred_type

        elif pred_type is None and batched:
            if self._batched_mjr:
                test_pred = pred[0] 
            else:
                test_pred = [p[0] if not isinstance(p,dict) else {k:v[0] for k,v in p.items()} for p in pred]
            self._determine_pred_format(test_pred,actions[0])
            pred_type = self._pred_type

        if not self._with_info:
            pred_info = {}
        elif self._info_dict:
            pred_info = {k:[p[-1][k] for p in pred] for k in pred[0][-1]} if self._batched_mjr else pred[-1]
        else:
            pred_info = {'_': [p[-1] for p in pred]} if batched else {'_': pred[-1]}

        if self._pred_type == 2:
            if not batched:
                pmf = pred[0] if pred_info else pred
                action,score = self._get_pmf_action_score(self._rng,pmf,actions)
            else:
                pmf = [p[0] if pred_info else p for p in pred] if self._batched_mjr else pred[0]
                action,score = list(zip(*map(self._get_pmf_action_score, repeat(self._rng), pmf, actions)))
        else:
            action,score = list(zip(*[p[:2] for p in pred])) if self._batched_mjr else pred[:2]

        if batched:
            return Batch(action), Batch(score), pred_info
        else:
            return action, score, pred_info

    def learn(self, context, actions, action, reward, probability, **kwargs) -> None:

        if self._batched is None:
            self._batched = isinstance(context,Batch)

        if self._batched and self._batched_lrn == False:
            for i in range(len(context)):
                self._safe_learn(context[i],actions[i],action[i],reward[i],probability[i], {k:v[i] for k,v in kwargs.items()})
        else:
            try:
                self._safe_learn(context,actions,action,reward,probability,kwargs)
            except:
                all_failed = False
                try:
                    for i in range(len(context)):
                        self._safe_learn(context[i],actions[i],action[i],reward[i],probability[i], {k:v[i] for k,v in kwargs.items()})
                    self._batched_lrn = False
                except:
                    all_failed = True
                if all_failed: raise

    def _determine_pred_type(self,pred,is_discrete) -> Optional[int]:
        if self._is_type_1(pred): raise CobaException("PDF predictions are currently not supported.")
        if self._is_type_2(pred,is_discrete): return 2
        if self._is_type_3(pred,is_discrete): return 3
        return None

    def _determine_has_info(self,pred,pred_type) -> Mapping[Any,Any]:
        if pred_type == 2:
            return not isinstance(pred[0],(int,float))
        else: #pred_type==3
            return len(pred) == 3

    def _determine_pred_format(self, pred, actions):

        is_discrete = 0 < len(actions) and len(actions) < float('inf')
        pred_type = self._determine_pred_type(pred, is_discrete) or self.get_inferred_type(pred, actions)
        with_info = self._determine_has_info(pred,pred_type)

        self._pred_type = pred_type
        self._with_info = with_info
        self._info_dict = with_info and isinstance(pred[-1],dict) 

    def _safe_predict(self, context, actions):

        batched = self._batched

        if batched is None: 
            batched = isinstance(context,Batch)
            self._batched = batched

        if not batched:
            return self._learner.predict(context,actions)

        if batched:
            batched_lrn = self._batched_lrn
            if batched_lrn == True:
                return self._learner.predict(context,actions)
            elif batched_lrn == False:
                return list(map(self._learner.predict,context,actions))
            elif batched_lrn is None:
                try:
                    batch_size = len(context)
                    pred = self._learner.predict(context,actions)
                    n_rows = len(pred)
                    n_cols = len(pred[0])

                    self._batched_lrn = True

                    if n_rows != n_cols:
                        self._batched_mjr = len(pred) == batch_size
                    else:
                        #The major order of pred is not determinable. So we
                        #now do a small "test" to determine the major order.
                        test_pred = self._learner.predict(Batch([context[0]]),Batch([actions[0]]))
                        self._batched_mjr = len(test_pred) == 1
                    return pred

                except Exception as e:
                    self._batched_lrn = False
                    self._batched_mjr = True
                    try:
                        return list(map(self._learner.predict,context,actions))
                    except:
                        pass
                    raise

    def _safe_learn(self,context,actions,action,reward,probability,kwargs):
        if self._learn_type==3:
            self._learner.learn(context, actions, action, reward, probability, **kwargs)
        elif self._learn_type==2:
            self._learner.learn(context, action, reward, probability, kwargs.get('_',kwargs))
        elif self._learn_type==1:
            self._learner.learn(context, action, reward, probability)
        else:
            all_failed = False
            try:
                self._learner.learn(context, actions, action, reward, probability, **kwargs)
                self._learn_type = 3
            except:
                try:
                    self._learner.learn(context, action, reward, probability, kwargs.get('_',kwargs))
                    self._learn_type = 2
                except:
                    try:
                        self._learner.learn(context, action, reward, probability)
                        self._learn_type = 1
                    except:
                        all_failed = True

                if all_failed: raise

    def _is_type_1(self, pred):
        #PDF
        return callable(pred) or callable(pred[0])

    def _is_type_2(self, pred, is_discrete:bool):
        #PMF

        explicit = isinstance(pred,Probs) or isinstance(pred[0],Probs)
        pmf_sans_info = is_discrete and isinstance(pred,abc.Sequence) and len(pred) > 3 or len(pred) == 3 and not isinstance(pred[2],dict)
        pmf_with_info = is_discrete and isinstance(pred,abc.Sequence) and len(pred) == 2 and isinstance(pred[0],abc.Sequence)
        
        return explicit or pmf_sans_info or pmf_with_info

    def _is_type_3(self, pred, is_discrete:bool):
        #Action Score
        explicit = isinstance(pred,ActionScore) or isinstance(pred[0],ActionScore)
        with_info = is_discrete and len(pred) == 3 and not isinstance(pred[2],(float,int))
        return explicit or with_info or not is_discrete

    def get_inferred_type(self,pred,actions) -> int:
        possible_types = []

        if self._possible_type_2(pred,actions): possible_types.append(2)
        if self._possible_type_3(pred,actions): possible_types.append(3)

        if len(possible_types) == 1:
            return possible_types[0]
        else:
            raise CobaException("We were unable to parse the given prediction format." 
                " This is likely because action features were returned for a discrete"
                " problem. When the action space is discrete, and you wish to directly"
                " return the selected action rather than a PMF, please provide the"
                " action index (i.e., actions.index(action)). Alternatively, this can"
                " also happen for two action problems. In this case we suggest using"
                " coba.learners.Probs or coba.learners.ActionScore to provide explicit"
                " type information (e.g., return coba.learners.Probs([1,0])).")

    def _pred_0_possible_pmf(self,pred,actions):
        possible_pmf = pred[0] if isinstance(pred[0],abc.Sequence) else pred
        return isclose(sum(possible_pmf), 1, abs_tol=.001) and len(possible_pmf) == len(actions)

    def _pred_0_possible_action(self,pred,actions):
        is_discrete = 0 < len(actions) and len(actions) < float('inf')
        return not is_discrete or pred[0] in list(range(len(actions)))

    def _possible_type_2(self,pred,actions):
        #PMF
        return self._pred_0_possible_pmf(pred,actions)

    def _possible_type_3(self,pred,actions):
        #action,score
        correct_shape = len(pred) in [2,3]        
        return self._pred_0_possible_action(pred,actions) and correct_shape

    def _get_pmf_action_score(self,rng,pmf,actions):
        assert len(pmf) == len(actions), "The learner returned an invalid number of probabilities for the actions"
        assert isclose(sum(pmf), 1, abs_tol=.001), "The learner returned a pmf which does not sum to one."
        return rng.choice(list(enumerate(pmf)), pmf)

    def __str__(self) -> str:
        return self.full_name

def requires_hashables(cls:Type[Learner]):

    def make_hashable(item):
        if isinstance(item,Dense): return HashableDense(item)
        if isinstance(item,Sparse): return HashableSparse(item)
        return item

    old_predict = cls.predict
    old_learn   = cls.learn

    def new_predict(self,c,A):
        return old_predict(self,make_hashable(c),list(map(make_hashable,A)))

    def new_learn(self,c,A,a,r,p,**kwargs):
        old_learn(self,make_hashable(c),list(map(make_hashable,A)), make_hashable(a),r,p,**kwargs)

    cls.predict = new_predict
    cls.learn   = new_learn

    return cls



import time
import pickle
import warnings
import copy

from math import isnan
from statistics import median, stdev, mode
from numbers import Number
from operator import eq, getitem, methodcaller, itemgetter
from collections import defaultdict, deque
from functools import lru_cache
from itertools import islice, chain, tee, compress, repeat
from typing import Optional, Sequence, Union, Iterable, Any, Tuple, Callable, Mapping
from coba.backports import Literal

from coba            import pipes, primitives
from coba.random     import CobaRandom
from coba.exceptions import CobaException
from coba.statistics import iqr
from coba.utilities  import peek_first
from coba.primitives import ScaleReward, BinaryReward, SequenceReward, BatchReward
from coba.primitives import Feedback, BatchFeedback, Categorical
from coba.learners   import Learner, SafeLearner

from coba.environments.primitives import EnvironmentFilter, Interaction

class Identity(pipes.Identity, EnvironmentFilter):
    """Return whatever interactions are given to the filter."""
    pass

class Take(pipes.Take, EnvironmentFilter):
    """Take a fixed number of interactions from an Environment."""
    pass

class Shuffle(pipes.Shuffle, EnvironmentFilter):
    """Shuffle a sequence of Interactions in an Environment."""
    pass

class Reservoir(pipes.Reservoir, EnvironmentFilter):
    """Take a fixed number of random Interactions from an Environment."""
    pass

class Cache(pipes.Cache, EnvironmentFilter):
    """Cache all interactions that come before this filter and use the cache in the future."""

    def filter(self, items: Iterable[Interaction]) -> Iterable[Interaction]:
        yield from map(methodcaller('copy'), super().filter(items))

class Scale(EnvironmentFilter):
    """Shift and scale features to precondition them before learning."""

    def __init__(self,
        shift: Union[Number,Literal["min","mean","med"]] = 0,
        scale: Union[Number,Literal["minmax","std","iqr","maxabs"]] = "minmax",
        target: Literal["context","rewards","argmax"] = "context",
        using: Optional[int] = None):
        """Instantiate a Scale filter.

        Args:
            shift: The statistic to use to shift each context feature.
            scale: The statistic to use to scale each context feature.
            target: The target data we wish to scale in the environment.
            using: The number of interactions to use when calculating the necessary statistics.
        """

        assert isinstance(shift,Number) or shift in ["min","mean","med"]
        assert isinstance(scale,Number) or scale in ["minmax","std","iqr","maxabs"]

        self._times = [0,0,0,0]

        self._shift  = shift
        self._scale  = scale
        self._using  = using
        self._target = target

    @property
    def params(self) -> Mapping[str, Any]:
        return {
            "scale_shift": self._shift,
            "scale_scale": self._scale,
            "scale_using": self._using,
            "scale_target": self._target
        }

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        first, interactions = peek_first(Mutable().filter(interactions))

        if not interactions: return []

        remaining_interactions = iter(interactions)
        fitting_interactions   = list(islice(remaining_interactions,self._using))

        first_context = first['context']
        first_actions = first.get('actions')

        if isinstance(first_context, primitives.Sparse) and self._target=="context" and self._shift != 0:
            raise CobaException("Shift is required to be 0 for sparse environments. Otherwise the environment will become dense.")

        is_discrete       = first_actions and len(first_actions) > 0
        is_dense_context  = isinstance(first_context, primitives.Dense)
        is_sparse_context = isinstance(first_context, primitives.Sparse)
        is_value_context  = not (is_dense_context or is_sparse_context)

        #get the values we wish to scale
        start = time.time()
        if self._target == "context" and is_dense_context:
            potential_cols = [i for i,v in enumerate(first['context']) if isinstance(v,(int,float))]
            if len(potential_cols) == 0:
                unscaled = []
            elif len(potential_cols) == 1:
                unscaled = [ tuple(map(itemgetter(*potential_cols),map(getitem,fitting_interactions,repeat("context")))) ]
            else:
                unscaled = list(zip(*map(itemgetter(*potential_cols),map(getitem,fitting_interactions,repeat("context")))))

        elif self._target == "context" and is_sparse_context:
            unscalable_cols = {k for k,v in first['context'].items() if not isinstance(v,(int,float))}
            unscaled = defaultdict(list)
            for interaction in fitting_interactions:
                context = interaction['context']
                for k in context.keys()-unscalable_cols:
                    unscaled[k].append(context[k])

        elif self._target == "context" and is_value_context:
            unscaled = [interaction['context'] for interaction in fitting_interactions]

        elif self._target == "rewards" and is_discrete:
            rwd_vals = lambda i: list(map(i['rewards'].eval,range(len(i['actions']))))
            unscaled = sum(list(map(rwd_vals,fitting_interactions)),[])

        elif self._target == "rewards" and not is_discrete:
            unscaled = []

        elif self._target == "argmax":
            unscaled = [interaction['rewards'].argmax() for interaction in fitting_interactions]
        self._times[1] += time.time()-start

        start = time.time()
        #determine the scale and shift values
        if self._target == "context" and is_dense_context:
            shifts_scales = []
            for i,col in zip(potential_cols,unscaled):
                shift_scale = self._get_shift_and_scale(col)
                if shift_scale is not None:
                    shifts_scales.append((i,)+shift_scale)

        elif self._target == "context" and is_sparse_context:
            shifts_scales = {}
            for k,col in unscaled.items():
                vals = col + [0]*(len(fitting_interactions)-len(col))
                shift_scale = self._get_shift_and_scale(vals)
                if shift_scale is not None:
                    shifts_scales[k] = shift_scale

        elif self._target in ['context','rewards','argmax']:
            shifts_scales = self._get_shift_and_scale(unscaled)
        self._times[2] += time.time()-start

        start = time.time()
        if not shifts_scales:
            yield from chain(fitting_interactions, remaining_interactions)

        #now scale return
        elif self._target == "context" and is_dense_context:
            for interaction in chain(fitting_interactions, remaining_interactions):
                context = interaction['context']
                for i,shift,scale in shifts_scales:
                     context[i] = (context[i]+shift)*scale
                yield interaction

        elif self._target == "context" and is_sparse_context:
            for interaction in chain(fitting_interactions, remaining_interactions):
                new = interaction # Mutable copies
                context = new['context']
                for k in shifts_scales.keys() & context.keys():
                    (shift,scale) = shifts_scales[k]
                    context[k] = (context[k]+shift)*scale
                yield new

        elif self._target == "context" and is_value_context:
            (shift,scale) = shifts_scales
            for interaction in chain(fitting_interactions, remaining_interactions):
                new = interaction.copy()
                new['context'] = (new['context']+shift)*scale
                yield new

        elif self._target == "rewards":
            (shift,scale) = shifts_scales
            for interaction in chain(fitting_interactions, remaining_interactions):
                new = interaction.copy()
                new['rewards'] = ScaleReward(new['rewards'], shift, scale, 'value')
                yield new

        elif self._target == "argmax":
            (shift,scale) = shifts_scales
            for interaction in chain(fitting_interactions, remaining_interactions):
                new = interaction.copy()
                new['rewards'] = ScaleReward(new['rewards'], shift, scale, 'argmax')
                yield new
        self._times[3] += time.time()-start

    def _get_shift_and_scale(self,values) -> Tuple[float,float]:
        try:
            shift = self._shift_value(values)
            scale = self._scale_value(values,shift)

            if isnan(shift):
                #this is a trick, nan != nan so equality will tell
                #if a value is not equal with itself then it is nan.
                #Using this trick is about 2x faster than using isnan.
                not_nan_vals = list(compress(values,map(eq,values,values)))
                shift = self._shift_value(not_nan_vals)
                scale = self._scale_value(not_nan_vals,shift)

            return shift,scale
        except TypeError:
            return None

    def _shift_value(self, values) -> float:
        shift = self._shift
        if shift == "min":
            return -min(values)
        elif shift == "mean":
            return -sum(values)/len(values) #mean() is very slow due to calculations for precision
        elif shift == "med":
            return -median(values)
        return shift

    def _scale_value(self, values, shift) -> float:
        scale     = self._scale
        scale_num = scale
        scale_den = 1

        if scale == "minmax":
            scale_num = 1
            scale_den = max(values)-min(values)
        elif scale == "std":
            scale_num = 1
            scale_den = stdev(values)
        elif scale == "iqr":
            scale_num = 1
            scale_den = iqr(values)
        elif scale == "maxabs":
            scale_num = 1
            scale_den = max([abs(v+shift) for v in values])

        return scale_num if scale_den < .000001 else scale_num/scale_den

class Impute(EnvironmentFilter):
    """Impute missing values (nan) in Interaction contexts."""

    def __init__(self,
        stat : Literal["mean","median","mode"] = "mean",
        indicator: bool = True,
        using: Optional[int] = None):
        """Instantiate an Impute filter.

        Args:
            stat: The statistic to use for impuatation.
            indicator: Indicates whether a new binary feature should be added for missingness.
            using: The number of interactions to use to calculate the imputation statistics.
        """

        assert stat in ["mean","median","mode"]

        self._stat  = stat
        self._miss  = indicator
        self._using = using
        self._times = [0,0,0,0]

    @property
    def params(self) -> Mapping[str, Any]:
        return { "impute_stat": self._stat, "impute_using": self._using, "impute_indicator":self._miss }

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        first, interactions = peek_first(Mutable().filter(interactions))

        if not interactions: return None

        is_dense  = isinstance(first['context'], primitives.Dense)
        is_sparse = isinstance(first['context'], primitives.Sparse)
        is_value  = not is_dense and not is_sparse

        interactions       = iter(interactions)
        using_interactions = list(islice(interactions,self._using))
        other_interactions = interactions

        start = time.time()
        if is_dense:
            if self._stat in ["mean","median"]:
                imputable_cols = [i for i,v in enumerate(first['context']) if isinstance(v,(int,float) or v is None)]
            else:
                imputable_cols = list(range(len(first['context'])))

        elif is_sparse:
            if self._stat in ['mean','median']:
                unimputable_cols = {k for k,v in first['context'].items() if not isinstance(v,(int,float) or v is None)}
            else:
                unimputable_cols = {}

        elif is_value:
            if self._stat in ["mean","median"]:
                imputable_cols = isinstance(first['context'],(int,float)) or (first['context'] is None)
            else:
                imputable_cols = True
        self._times[0] += time.time()-start

        start = time.time()
        #get unimputed values
        if is_dense:
            if len(imputable_cols) == 0:
                unimputed = []
            elif len(imputable_cols) == 1:
                unimputed = [ tuple(map(itemgetter(*imputable_cols),map(getitem,using_interactions,repeat("context")))) ]
            else:
                unimputed = list(zip(*map(itemgetter(*imputable_cols),map(getitem,using_interactions,repeat("context")))))

        elif is_sparse:
            unimputed = defaultdict(list)
            for interaction in using_interactions:
                context = interaction['context']
                for k in context.keys()-unimputable_cols:
                    unimputed[k].append(context[k])

        elif is_value:
            unimputed = [ interaction['context'] for interaction in using_interactions ]
        self._times[1] += time.time()-start

        start = time.time()
        #calculate imputation statistics
        if is_dense:
            imputations = {}
            impute_binary = {}
            for i, col in zip(imputable_cols,unimputed):
                imputation = self._get_imputation(col)
                if imputation is not None:
                    imputations[i] = imputation
                    if self._miss and any([c is None for c in col]): 
                        impute_binary[i] = len(impute_binary)

        elif is_sparse:
            imputations = {}
            impute_binary = {}
            binary_template = {}
            for k,col in unimputed.items():
                imputation = self._get_imputation(col + [0]*(len(using_interactions)-len(col)))
                if imputation is not None:
                    imputations[k] = imputation
                    if self._miss and any([c is None for c in col]):
                        impute_binary[k] = f"{k}_is_missing"
                        binary_template[f"{k}_is_missing"] = 0

        elif is_value:
            imputations = self._get_imputation(unimputed)
            impute_binary = self._miss and any([c is None for c in unimputed])
        self._times[2] += time.time()-start

        start = time.time()
        for interaction in chain(using_interactions,other_interactions):

            context = interaction['context']

            if is_dense:
                is_missing = [0]*len(impute_binary)
                for k,v in enumerate(context):
                    if v is None and k in imputations: 
                        context[k] = imputations[k]
                        if k in impute_binary: 
                            is_missing[impute_binary[k]] = 1
                context += is_missing

            elif is_sparse:

                is_missing = binary_template.copy()
                for k,v in context.items():
                    if v is None: 
                        context[k] = imputations[k]
                        if k in impute_binary:
                            is_missing[impute_binary[k]] = 1
                context.update(is_missing)

            elif is_value:
                if impute_binary:
                    if context is None: 
                        interaction["context"] = [imputations,1]
                    else:
                        interaction["context"] = [context,0]
                else:
                    if context is None:
                        interaction["context"] = imputations

            yield interaction
        self._times[3] += time.time()-start

    def _get_imputation(self,values):
        try:
            values = [v for v in values if v is not None]
            if self._stat == "mean":
                return sum(values)/len(values)
            if self._stat == "median":
                return median(values)
            if self._stat == "mode":
                return mode(values)
        except:
            return None

class Sparse(EnvironmentFilter):
    """Sparsify an environment's feature representation.

    This has little utility beyond debugging.
    """

    def __init__(self, context:bool = True, action:bool = False):
        """Instantiate a Sparse filter.

        Args:
            context: If True then contexts should be made sparse otherwise leave them alone.
            action: If True then actions should be made sparse otherwise leave them alone.
        """

        self._context = context
        self._action  = action

    @property
    def params(self) -> Mapping[str, Any]:
        return { "sparse_C": self._context, "sparse_A": self._action }

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        for interaction in interactions:

            new = interaction.copy()            

            if self._context:
                new['context'] = self._make_sparse(new['context'])

            if self._action and 'actions' in new:
                new['actions'] = list(map(self._make_sparse,new['actions']))

            if self._action and 'action' in new:
                new['action'] = self._make_sparse(new['action'])

            yield new

    def _make_sparse(self, value) -> Optional[dict]:
        if value is None:
            return value
        if isinstance(value,primitives.Dense):
            return dict(enumerate(value))
        if isinstance(value,primitives.Sparse):
            return value
        return {0:value}

class Cycle(EnvironmentFilter):
    """Cycle all rewards associated with actions by one place.

    This filter is useful for testing an algorithms response to a non-stationary shock.
    """

    def __init__(self, after:int = 0):
        """Instantiate a Cycle filter.

        Args:
            after: How many interactions should be seen before applying the cycle filter.
        """
        self._after = after

    @property
    def params(self) -> Mapping[str, Any]:
        return { "cycle_after": self._after }

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        underlying_iterable     = iter(interactions)
        sans_cycle_interactions = islice(underlying_iterable, self._after)
        with_cycle_interactions = underlying_iterable

        yield from sans_cycle_interactions

        first, with_cycle_interactions = peek_first(with_cycle_interactions)

        if with_cycle_interactions:

            action_set = set(first['actions'])
            n_actions  = len(action_set)

            onehot_actions  = {tuple([0]*n+[1]+[0]*(n_actions-n-1)) for n in range(n_actions)}
            is_discrete     = 0 < n_actions and n_actions < float('inf')
            is_onehots      = action_set == onehot_actions
            is_categoricals = all([isinstance(a,str) for a in action_set])

            if not is_discrete or (not is_onehots and not is_categoricals):
                warnings.warn("Cycle only works for discrete environments without action features. It will be ignored in this case.")
                yield from with_cycle_interactions
            else:
                for interaction in with_cycle_interactions:
                    rewards = deque(interaction['rewards'])
                    rewards.rotate(1)

                    new = interaction.copy()
                    new['rewards'] = SequenceReward(list(rewards))

                    yield new

class Flatten(EnvironmentFilter):
    """Flatten the context and action features for interactions."""

    def __init__(self):
        self._flattener = pipes.Flatten()

    @property
    def params(self) -> Mapping[str, Any]:
        return { "flat": True }

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        I1,I2,I3 = tee(interactions,3)

        interactions      = I1
        flat_context_iter = self._flattener.filter(i['context'] for i in I2                   )
        flat_actions_iter = self._flattener.filter(a            for i in I3 for a in i['actions'])

        for interaction in interactions:
            new = interaction.copy()
            new['context'] = next(flat_context_iter)
            new['actions'] = list(islice(flat_actions_iter,len(interaction['actions'])))
            yield new

class Binary(EnvironmentFilter):
    """Binarize all rewards to either 1 (max rewards) or 0 (all others)."""

    @property
    def params(self) -> Mapping[str, Any]:
        return { "binary": True }

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
        for interaction in interactions:
            new = interaction.copy()
            new['rewards'] = BinaryReward(interaction['rewards'].argmax())
            yield new

class Sort(EnvironmentFilter):
    """Sort a sequence of Interactions in an Environment."""

    def __init__(self, *keys: Union[str,int,Sequence[Union[str,int]]]) -> None:
        """Instantiate a Sort filter.

        Args:
            *keys: The context items that should be sorted on.
        """

        self._keys = list(pipes.Flatten().filter([list(keys)]))[0]

    @property
    def params(self) -> Mapping[str, Any]:
        return { "sort": self._keys or '*' }

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        full_sorter = lambda interaction: tuple(interaction['context']                                 )
        list_sorter = lambda interaction: tuple(interaction['context'][key]       for key in self._keys)
        dict_sorter = lambda interaction: tuple(interaction['context'].get(key,0) for key in self._keys)

        first, interactions = peek_first(interactions)
        is_sparse           = isinstance(first['context'],primitives.Sparse)

        sorter = full_sorter if not self._keys else dict_sorter if is_sparse else list_sorter

        return sorted(interactions, key=sorter)

class Where(EnvironmentFilter):
    """Define Environment selection criteria for an Environments pipe."""

    def __init__(self, *, n_interactions: Union[int,Tuple[Optional[int],Optional[int]]] = None) -> None:
        """Instantiate a Where filter.

        Args:
            n_interactions: The minimum, maximum or exact number of interactions Environments must have.
        """

        self._n_interactions = n_interactions

    @property
    def params(self) -> Mapping[str, Any]:
        params = {}

        if self._n_interactions is not None:
            params["where_n_interactions"] = self._n_interactions

        return params

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        interactions = iter(interactions)

        if self._n_interactions is None or self._n_interactions == (None,None):
            min_interactions  = None
            max_interactions  = None
            take_interactions = 0
        elif isinstance(self._n_interactions, int):
            min_interactions  = self._n_interactions
            max_interactions  = self._n_interactions
            take_interactions = self._n_interactions+1
        else:
            min_interactions  = self._n_interactions[0]
            max_interactions  = self._n_interactions[1]
            take_interactions = max(filter(lambda x: x is not None, list(self._n_interactions)))+1

        taken_interactions = list(islice(interactions, take_interactions))

        if max_interactions is not None and len(taken_interactions) > max_interactions:
            return []

        if min_interactions is not None and len(taken_interactions) < min_interactions:
            return []

        return chain(taken_interactions, interactions)

class Riffle(EnvironmentFilter):
    """Riffle shuffle Interactions by taking actions from the end and evenly distributing into the beginning."""

    def __init__(self, spacing: int = 3, seed=1) -> None:
        """Instantiate a Riffle filter.

        Args:
            spacing: The number of interactions from the beginning between each interaction shuffled in from the end.
            seed: The seed used to determine the location of each ending interaction when placed within its beginning space.
        """
        self._spacing = spacing
        self._seed = seed

    @property
    def params(self) -> Mapping[str, Any]:
        return {"riffle_spacing": self._spacing, "riffle_seed": self._seed}

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        rng          = CobaRandom(self._seed)
        interactions = list(interactions)

        for i in range(int(len(interactions)/(self._spacing+1))):
            interactions.insert(i*self._spacing+rng.randint(0,self._spacing), interactions.pop())

        return interactions

class Noise(EnvironmentFilter):
    """Introduce noise to an environment."""

    def __init__(self,
        context: Callable[[float,CobaRandom], float] = None,
        action : Callable[[float,CobaRandom], float] = None,
        reward : Callable[[float,CobaRandom], float] = None,
        seed   : int = 1) -> None:
        """Instantiate a Noise EnvironmentFilter.

        Args:
            context: A noise generator for context features.
            action : A noise generator for action features.
            reward : A noise generator for rewards.
            seed   : The seed initializing the random state of the noise generators.
        """

        self._args = (context,action,reward,seed)
        self._no_noise = lambda x, _: x

        if context is None and action is None and reward is None:
            context = lambda x, rng: x+rng.gauss(0,1)

        self._context_noise = context or self._no_noise
        self._action_noise  = action  or self._no_noise
        self._reward_noise  = reward  or self._no_noise
        self._seed          = seed

    def __reduce__(self) -> tuple:
        try:
            pickle.dumps(self._args)
        except Exception:
            message = (
                "We were unable to pickle the Noise filter. This is likely due to using lambda functions for noise generation. "
                "To work around this we recommend you first define your lambda functions as a named function and then pass the "
                "named function to Noise."
            )
            raise CobaException(message)
        else:
            return (Noise, self._args)

    @property
    def params(self) -> Mapping[str, Any]:

        params = {}

        if self._context_noise != self._no_noise: params['context_noise'] = True
        if self._action_noise != self._no_noise : params['action_noise' ] = True
        if self._reward_noise != self._no_noise : params['reward_noise' ] = True

        params['noise_seed'] = self._seed

        return params

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        rng = CobaRandom(self._seed)

        for interaction in interactions:
            new = interaction.copy()

            noisy_context = self._noises(interaction['context'], rng, self._context_noise)
            noisy_actions = [self._noises(a, rng, self._action_noise) for a in interaction['actions'] ]
            noisy_rewards = [ self._noises(r, rng, self._reward_noise) for r in interaction['rewards'] ]
            noisy_rewards = SequenceReward(noisy_rewards)

            new['context'] = noisy_context
            new['actions'] = noisy_actions
            new['rewards'] = noisy_rewards

            yield new

    def _noises(self, value:Union[None,float,str,Mapping,Sequence], rng: CobaRandom, noiser: Callable[[float,CobaRandom], float]):

        if isinstance(value, primitives.Sparse):
            #we sort so that noise generation is deterministic with respect to seed
            return { k:self._noise(v, rng, noiser) for k,v in sorted(value.items()) }

        if isinstance(value, primitives.Dense):
            return [ self._noise(v, rng, noiser) for v in value ]

        return self._noise(value, rng, noiser)

    def _noise(self, value:Union[None,float,str], rng: CobaRandom, noiser: Callable[[float,CobaRandom], float]) -> float:

        return value if not isinstance(value,(int,float)) else noiser(value, rng)

class Params(EnvironmentFilter):
    """Add parameters to an environment."""
    def __init__(self, params: Mapping[str, Any]) -> None:
        self._params = params

    @property
    def params(self) -> Mapping[str, Any]:
        return self._params

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
        return interactions

class Grounded(EnvironmentFilter):

    class GroundedFeedback(Feedback):
        def __init__(self, goods, bads, argmax, seed):
            self._rng    = None
            self._goods  = goods
            self._bads   = bads
            self._seed   = seed
            self._argmax = argmax

        @lru_cache(maxsize=None)
        def eval(self, arg):
            if not self._rng: self._rng = CobaRandom(self._seed)
            if arg == self._argmax:
                return self._rng.choice(self._goods) 
            else:
                return self._rng.choice(self._bads) 

    def __init__(self, n_users: int, n_normal:int, n_words:int, n_good:int, seed:int) -> None:
        self._n_users  = self._cast_int(n_users , "n_users" )
        self._n_normal = self._cast_int(n_normal, "n_normal")
        self._n_words  = self._cast_int(n_words , "n_words" )
        self._n_good   = self._cast_int(n_good  , "n_good"  )
        self._seed     = seed

        if n_normal > n_users:
            raise CobaException("Igl conversion can't have more normal users (n_normal) than total users (n_users).")

        if n_good > n_words:
            raise CobaException("Igl conversion can't have more good words (n_good) than total words (n_words).")

        self.userids   = list(range(self._n_users))
        self.normalids = self.userids[:self._n_normal]
        self.wordids   = list(range(self._n_words))
        self.goodwords = self.wordids[:self._n_good]
        self.badwords  = self.wordids[self._n_good:]

    @property
    def params(self) -> Mapping[str, Any]:
        return {
            "n_users"  : self._n_users,
            "n_normal" : self._n_normal,
            "n_good"   : self._n_good,
            "n_words"  : self._n_words,
            "igl_seed" : self._seed
        }

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
        rng = CobaRandom(self._seed)

        #we make it a set for faster contains checks
        normalids       = set(self.normalids) 
        normal          = [u in normalids for u in self.userids]
        userid_isnormal = list(zip(self.userids,normal))

        first,interactions = peek_first(interactions)

        if not interactions: return []

        is_binary_rwd = {0,1} == set(first['rewards'])
        first_context = first['context']

        is_sparse = isinstance(first_context,primitives.Sparse)
        is_dense  = isinstance(first_context,primitives.Dense)

        goods = [(g,) for g in self.goodwords]
        bads  = [(b,) for b in self.badwords ]

        for seed, interaction in enumerate(interactions, self._seed):

            userid,normal = rng.choice(userid_isnormal)
            argmax        = interaction['rewards'].argmax()

            new = interaction.copy()

            if not is_binary_rwd:
                new['rewards'] = BinaryReward(argmax)

            if normal:
                new['feedbacks'] = Grounded.GroundedFeedback(goods,bads,argmax,seed)
            else:
                new['feedbacks'] = Grounded.GroundedFeedback(bads,goods,argmax,seed)

            if is_sparse:
                new['context'] = dict(userid=userid,**new['context'])
            elif is_dense:
                new['context'] = (userid,)+tuple(new['context'])
            else:
                new['context'] = (userid, new['context'])

            new['userid'  ] = userid
            new['isnormal'] = normal

            yield new

    def _cast_int(self, value:Union[float,int], value_name:str) -> int:
        if isinstance(value, int): return value
        if isinstance(value, float) and value.is_integer(): return int(value)
        raise CobaException(f"{value_name} must be a whole number and not {value}.")

class Repr(EnvironmentFilter):
    def __init__(self, 
        cat_context:Literal["onehot","onehot_tuple","string"] = None,
        cat_actions:Literal["onehot","onehot_tuple","string"] = None) -> None:
        self._cat_context = cat_context
        self._cat_actions = cat_actions

    @property
    def params(self) -> Mapping[str, Any]:
        return {"cat_context": self._cat_context, "cat_actions": self._cat_actions}

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        first, interactions = peek_first(interactions)

        if not interactions: return []

        has_actions = 'actions' in first

        I = tee(interactions, 3 if has_actions else 2)

        interactions     = I[0]
        cat_context_iter = pipes.EncodeCatRows(self._cat_context).filter(i['context'] for i in I[1])

        if has_actions:
            cat_actions_iter = pipes.EncodeCatRows(self._cat_actions, True).filter(action for i in I[2] for action in i['actions'])

        for interaction in interactions:
            new = interaction.copy()
            new['context'] = next(cat_context_iter)

            if has_actions:
                new['actions'] = list(islice(cat_actions_iter,len(interaction['actions'])))

            yield new

class Batch(EnvironmentFilter):

    def __init__(self, batch_size: int) -> None:
        self._batch_size = batch_size

    @property
    def params(self) -> Mapping[str,Any]:
        return {'batched': self._batch_size}

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        for batch in self._batched(interactions, self._batch_size):
            new = { k: primitives.Batch(i[k] for i in batch) for k in batch[0] }
            if 'rewards' in new: new['rewards'] = BatchReward(new['rewards'])
            if 'feedbacks' in new: new['feedbacks'] = BatchFeedback(new['feedbacks'])
            yield new

    def _batched(self, iterable, n):
        "Batch data into lists of length n. The last batch may be shorter."
        #taken from python itertools recipes
        it = iter(iterable)
        batch = list(islice(it, n))
        while (batch):
            yield batch
            batch = list(islice(it, n))

class Unbatch(EnvironmentFilter):

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        first, interactions = peek_first(interactions)
        if not interactions: return []
        is_batched = isinstance(first[list(first.keys())[0]], primitives.Batch)

        if not is_batched:
            yield from interactions
        else:
            yield from self._unbatch(interactions)

    def _unbatch(self, interactions: Iterable[Interaction]):
        for interaction in interactions:
            batch_size = len(interaction[list(interaction.keys())[0]])
            for i in range(batch_size):
                yield { k: interaction[k][i] for k in interaction }

class BatchSafe(EnvironmentFilter):

    def __init__(self, filter: EnvironmentFilter) -> None:
        self._filter = filter

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
        first, interactions = peek_first(interactions)

        if not interactions: return []

        is_batched = isinstance(first[list(first.keys())[0]], primitives.Batch)

        if not is_batched:
            yield from self._filter.filter(interactions)
        else:
            batch_size = len(first[list(first.keys())[0]])
            unbatched = Unbatch().filter(interactions)
            filtered  = self._filter.filter(unbatched)
            rebatched = Batch(batch_size).filter(filtered)
            yield from rebatched

class Finalize(EnvironmentFilter):

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        first, interactions = peek_first(interactions)

        if not interactions: return []

        first_has_context = 'context' in first
        first_has_actions = 'actions' in first and first['actions']
        first_has_action  = 'action'  in first

        context_materialized = not first_has_context or isinstance(first['context']   ,(list,tuple,dict))
        actions_materialized = not first_has_actions or isinstance(first['actions'][0],(list,tuple,dict))
        action_materialized  = not first_has_action  or isinstance(first['action' ]   ,(list,tuple,dict))

        if first_has_context:
            is_dense_context  = isinstance(first['context'],primitives.Dense)
            is_sparse_context = isinstance(first['context'],primitives.Sparse)
        if first_has_actions:
            is_dense_action  = isinstance(first['actions'][0],primitives.Dense)
            is_sparse_action = isinstance(first['actions'][0],primitives.Sparse)
        elif first_has_action:
            is_dense_action  = isinstance(first['action'],primitives.Dense)
            is_sparse_action = isinstance(first['action'],primitives.Sparse)

        for interaction in Repr("onehot","onehot").filter(interactions):

            new = interaction.copy()

            if not context_materialized and is_dense_context:
                new['context'] = list(new['context'])
            elif not context_materialized and is_sparse_context :
                    new['context'] = new['context'].copy()

            if not actions_materialized and is_dense_action:
                new['actions'] = list(map(list,new['actions']))
            elif not actions_materialized and is_sparse_action:
                new['actions'] = list(map(methodcaller('copy'),new['actions']))

            if not action_materialized and is_dense_action:
                new['action'] = list(new['action'])
            elif not action_materialized and is_sparse_action:
                new['action'] = new['action'].copy()

            yield new

class Chunk(EnvironmentFilter):
    """A placeholder filter that exists only to semantically indicate how an environment pipe should be chunked for processing."""
    def filter(self, items: Iterable[Interaction]) -> Iterable[Interaction]:
        return items

class Logged(EnvironmentFilter):

    def __init__(self, learner: Learner) -> None:
        self._learner = learner

    @property
    def params(self) -> Mapping[str, Any]:
        return {"learner": SafeLearner(self._learner).params, "logged":True}

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        learner = SafeLearner(copy.deepcopy(self._learner))

        first, interactions = peek_first(interactions)

        predict = learner.predict
        learn   = learner.learn

        if 'context' not in first or 'actions' not in first or 'rewards' not in first:
            raise CobaException("We were unable to create a logged representation of the interaction.")

        for interaction in interactions:

            context = interaction['context']
            actions = interaction['actions']
            rewards = interaction['rewards']

            action,prob,info = predict(context, actions)

            reward = rewards.eval(action)

            learn(context, actions, action, reward, prob, **info)

            new = interaction.copy()

            new['action']      = action
            new['probability'] = prob
            new['reward']      = reward

            yield new

class Mutable(EnvironmentFilter):

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
        
        first, interactions = peek_first(interactions)
        
        if not interactions: return []

        first_is_dense   = isinstance(first['context'], primitives.Dense)
        first_is_sparse  = isinstance(first['context'], primitives.Sparse)
        first_is_value   = not (first_is_dense or first_is_sparse)
        first_is_mutable = isinstance(first['context'], (list,dict))

        if first_is_mutable:
            for interaction in interactions:
                new = interaction.copy()
                new['context'] = new['context'].copy()
                yield new
                            
        elif first_is_dense:
            for interaction in interactions:
                new = interaction.copy()
                new['context'] = list(new['context'])
                yield new
        
        elif first_is_sparse:
            for interaction in interactions:
                new = interaction.copy()
                new['context'] = dict(new['context'].items())
                yield new

        elif first_is_value:
            for interaction in interactions:
                yield interaction.copy()



from copy import deepcopy
from itertools import islice
from collections import defaultdict, Counter
from typing import Iterable, Sequence, Any, Optional, Union, Tuple

from coba.pipes import Source, Filter, SourceFilters
from coba.learners import Learner
from coba.contexts import CobaContext
from coba.environments import Environment, Finalize, BatchSafe, Chunk
from coba.utilities import peek_first

from coba.experiments.tasks import LearnerTask, EnvironmentTask, EvaluationTask
from coba.experiments.results import Result

class WorkItem:

    def __init__(self,
        env_id: Optional[int],
        lrn_id: Optional[int],
        env   : Optional[Environment],
        lrn   : Optional[Learner],
        task  : Union[LearnerTask, EnvironmentTask, EvaluationTask]) -> None:

        self.env_id = env_id
        self.lrn_id = lrn_id
        self.env    = env
        self.lrn    = lrn
        self.task   = task

class CreateWorkItems(Source[Iterable[WorkItem]]):

    def __init__(self,
        evaluation_pairs: Sequence[Tuple[Learner,Environment]],
        learner_task    : LearnerTask,
        environment_task: EnvironmentTask,
        evaluation_task : EvaluationTask) -> None:

        self._evaluation_pairs = evaluation_pairs
        self._environment_task = environment_task
        self._learner_task     = learner_task
        self._evaluation_task  = evaluation_task

    def read(self) -> Iterable[WorkItem]:

        #we rely on ids to make sure we don't do duplicate work. So long as self.evaluation_pairs
        #is always in the exact same order we should be fine. In the future we may want to consider.
        #adding a better check for environments other than assigning an index based on their order.

        lrns = dict()
        envs = dict()

        for lrn,env in self._evaluation_pairs:

            if lrn not in lrns:
                lrns[lrn] = len(lrns)
                yield WorkItem(None, lrns[lrn], None, lrn, self._learner_task)

            if env not in envs:
                envs[env] = len(envs)
                yield WorkItem(envs[env], None, env, None, self._environment_task)

            yield WorkItem(envs[env], lrns[lrn], env, lrn, self._evaluation_task) 

class RemoveFinished(Filter[Iterable[WorkItem], Iterable[WorkItem]]):
    def __init__(self, restored: Optional[Result]) -> None:
        self._restored = restored

    def filter(self, tasks: Iterable[WorkItem]) -> Iterable[WorkItem]:

        for task in tasks:

            is_learner_task = task.env_id is None
            is_environ_task = task.lrn_id is None
            is_eval_task    = not (is_learner_task or is_environ_task)

            if not self._restored:
                yield task
            elif is_learner_task and task.lrn_id not in self._restored.learners:
                yield task
            elif is_environ_task and task.env_id not in self._restored.environments:
                yield task
            elif is_eval_task and (task.env_id, task.lrn_id) not in self._restored._interactions:
                yield task

class ChunkByChunk(Filter[Iterable[WorkItem], Iterable[Sequence[WorkItem]]]):

    def filter(self, items: Iterable[WorkItem]) -> Iterable[Sequence[WorkItem]]:

        items  = list(items)
        chunks = defaultdict(list)

        workitems_sans_env = [t for t in items if t.env_id is None    ]
        workitems_with_env = [t for t in items if t.env_id is not None]

        for env_item in workitems_with_env:
            chunks[self._get_last_chunk(env_item.env)].append(env_item)

        for lrn_item in workitems_sans_env:
            yield [lrn_item]

        for workitem in chunks.pop(None,[]):
            yield [workitem]

        for chunk in sorted(chunks.values(), key=lambda chunk: min([c.env_id for c in chunk])):
            yield list(sorted(chunk, key=lambda c: (c.env_id, -1 if c.lrn_id is None else c.lrn_id)))

    def _get_last_chunk(self, env):
        if isinstance(env, SourceFilters):
            for pipe in reversed(list(env)):
                if isinstance(pipe, Chunk):
                    return pipe
        return None

class MaxChunkSize(Filter[Iterable[Sequence[WorkItem]], Iterable[Sequence[WorkItem]]]):
    def __init__(self, max_tasks) -> None:
        self._max_tasks = max_tasks

    def filter(self, chunks: Iterable[Sequence[WorkItem]]) -> Iterable[Sequence[WorkItem]]:

        for chunk in chunks:
            chunk = iter(chunk)
            max_task_chunk = list(islice(chunk,self._max_tasks or None))
            while max_task_chunk:
                yield max_task_chunk
                max_task_chunk = list(islice(chunk,self._max_tasks or None))

class ProcessWorkItems(Filter[Iterable[WorkItem], Iterable[Any]]):

    def filter(self, chunk: Iterable[WorkItem]) -> Iterable[Any]:

        chunk = list(chunk)
        empty_envs = set()

        finalizer = BatchSafe(Finalize())

        if not chunk: return

        if len(chunk) > 1:
            #We sort in case there are multiple chunks in the pipe. Sorting means we can free chunks from memory as we go.
            chunk = sorted(chunk, key=lambda item: self._env_ids(item)+self._lrn_ids(item))
            chunk = list(reversed(chunk))

        learner_eval_counts = Counter([item.lrn_id for item in chunk if item.lrn and item.env])

        with CobaContext.logger.log(f"Processing chunk..."):

            while chunk:
                try:
                    item = chunk.pop()

                    if item.env is None:
                        with CobaContext.logger.time(f"Recording Learner {item.lrn_id} parameters..."):
                            row = item.task.process(item.lrn)
                            yield ["T1", item.lrn_id, row]

                    if item.lrn is None:
                        with CobaContext.logger.time(f"Recording Environment {item.env_id} statistics..."):
                            row = item.task.process(item.env,finalizer.filter(item.env.read()))
                            yield ["T2", item.env_id, row]

                    if item.env and item.lrn and item.env_id not in empty_envs:

                        with CobaContext.logger.time(f"Peeking at Environment {item.env_id}..."):
                            interactions = peek_first(item.env.read())[1]

                        if not interactions: 
                            CobaContext.logger.log(f"Environment {item.env_id} has nothing to evaluate (this is likely due to having too few interactions).")
                            empty_envs.add(item.env_id)
                            continue

                        with CobaContext.logger.time(f"Evaluating Learner {item.lrn_id} on Environment {item.env_id}..."):
                            lrn = item.lrn if learner_eval_counts[item.lrn_id] == 1 else deepcopy(item.lrn)
                            row = list(item.task.process(lrn, finalizer.filter(interactions)))
                            yield ["T3", (item.env_id, item.lrn_id), row]

                except Exception as e:
                    CobaContext.logger.log(e)

    # def _cache_ids(self, item: WorkItem) -> Tuple[int,...]:
        #I'm not sure this is necessary and it makes sorting by env-ids difficult which 
        #isn't necessary but is nice to have when you're running experiments on a single 
        #process. Therefore, we're not going to use this code for now.
    #   return tuple(id(pipe) for pipe in reversed(list(item.env)) if isinstance(pipe,Cache) ) if isinstance(item.env, SourceFilters) else ()

    def _env_ids(self, item: WorkItem):
        return (-1,) if item.env_id is None else (item.env_id,)

    def _lrn_ids(self, item: WorkItem):
        return (-1,) if item.lrn_id is None else (item.lrn_id,)
 


from queue import Queue
from typing import Iterable, Any, Union, Dict

from coba.exceptions import CobaException

from coba.pipes.sinks      import QueueSink
from coba.pipes.sources    import QueueSource
from coba.pipes.primitives import Filter, Source, Sink

class SourceFilters(Source):
    def __init__(self, *pipes: Union[Source,Filter]) -> None:
        if isinstance(pipes[0], SourceFilters):
            self._source = pipes[0]._source
            self._filter = FiltersFilter(*pipes[0]._filter._filters, *pipes[1:])
        else:
            self._source = pipes[0]
            self._filter = FiltersFilter(*pipes[1:])

    @property
    def params(self) -> Dict[str,Any]:

        try:
            source_params = self._source.params
        except:
            source_params = {}

        return { **source_params, **self._filter.params }

    def read(self) -> Any:
        return self._filter.filter(self._source.read())

    def __str__(self) -> str:
        return ",".join(map(str,[self._source, self._filter]))

    def __getitem__(self, index:int) -> Union[Source,Filter]:
        if index == 0 or index == -len(self):
            return self._source
        elif index < 0:
            return self._filter[index]
        else:
            return self._filter[index-1]

    def __iter__(self) -> Iterable[Union[Source,Filter]]:
        yield self._source
        yield from self._filter

    def __len__(self) -> int:
        return len(self._filter)+1

class FiltersFilter(Filter):

    def __init__(self, *pipes: Filter):
        self._filters = sum([f._filters if isinstance(f, FiltersFilter) else [f] for f in pipes ],[])

    @property
    def params(self) -> Dict[str,Any]:
        return { k:v for f in self._filters if hasattr(f,'params') for k,v in f.params.items() }

    def filter(self, items: Any) -> Any:
        for filter in self._filters:
            items = filter.filter(items)
        return items

    def __str__(self) -> str:
        return ",".join(map(str,self._filters))

    def __getitem__(self, index: int) -> Filter:
        return self._filters[index]

    def __iter__(self) -> Iterable[Filter]:
        return iter(self._filters)
    
    def __len__(self) -> int:
        return len(self._filters)

class FiltersSink(Sink):

    def __init__(self, *pipes: Union[Filter,Sink]) -> None:

        filters = list(pipes[:-1])
        sink    = pipes[-1 ]

        if isinstance(sink, FiltersSink):
            filters += sink._filter._filters
            sink     = sink._sink

        self._filter = FiltersFilter(*filters)
        self._sink   = sink

    @property
    def params(self) -> Dict[str,Any]:

        try:
            sink_params = self._sink.params
        except:
            sink_params = {}

        return { **self._filter.params, **sink_params }

    def write(self, items: Iterable[Any]):
        self._sink.write(self._filter.filter(items))

    def __str__(self) -> str:
        return ",".join(map(str,[self._filter, self._sink]))

    def __getitem__(self, index: int) -> Union[Filter,Sink]:
        if index == -1 or index == len(self._filter):
            return self._sink
        elif index < 0:
            return self._filter[index+1] 
        else:
            return self._filter[index]
    
    def __iter__(self) -> Iterable[Filter]:
        yield from self._filter
        yield self._sink

    def __len__(self) -> int:
        return len(self._filter)+1

class Pipes:
    """A helper class to compose sequences of pipes."""
    class Line:

        def __init__(self, *pipes: Union[Source,Filter,Sink]) -> None:
            self.pipes = list(pipes)

        def run(self) -> None:
            """Run the pipeline."""

            source  = self.pipes[0   ]
            filters = self.pipes[1:-1]
            sink    = self.pipes[-1  ]

            item = source.read()

            for filter in filters:
                item = filter.filter(item)

            sink.write(item)

        @property
        def params(self) -> Dict[str, Any]:
            return { k:v for p in self.pipes if hasattr(p,'params') for k,v in p.params.items() }

        def __str__(self) -> str:
            return ",".join(filter(None,map(str,self.pipes)))

        def __len__(self) -> int:
            return len(self.pipes)
        
        def __getitem__(self, index:int) -> Union[Source,Filter,Sink]:
            return self.pipes[index]

    @staticmethod
    def join(*pipes: Union[Source, Filter, Sink]) -> Union[Source, Filter, Sink, Line]:
        """Join a sequence of pipes into a single pipe.

        Args:
            pipes: a sequence of pipes.

        Returns:
            A single pipe that is a composition of the given pipes. The type of pipe returned
            is determined by the sequence given. A sequence of Filters will return a Filter. A
            sequence that begins with a Source and is followed by Filters will return a Source.
            A sequence that starts with Filters and ends with a Sink will return a Sink. A
            sequence that begins with a Source and ends with a Sink will return a completed pipe.
        """
        if len(pipes) == 0:
            raise CobaException("No pipes were passed to join.")

        if len(pipes) == 1 and any(hasattr(pipes[0],attr) for attr in ['read','filter','write']):
            return pipes[0]

        first = pipes[0 ] if not isinstance(pipes[0 ], Foreach) else pipes[0 ]._pipe
        last  = pipes[-1] if not isinstance(pipes[-1], Foreach) else pipes[-1]._pipe

        if hasattr(first,'read') and hasattr(last,'write'):
            return Pipes.Line(*pipes)

        if hasattr(first,'read') and hasattr(last,'filter'):
            return SourceFilters(*pipes)

        if hasattr(first,'filter') and hasattr(last,'filter'):
            return FiltersFilter(*pipes)

        if hasattr(first,'filter') and hasattr(last,'write'):
            return FiltersSink(*pipes)

        raise CobaException("An unknown pipe was passed to join.")

class Foreach(Filter[Iterable[Any], Iterable[Any]], Sink[Iterable[Any]]):
    """A pipe that wraps an inner pipe and passes items to it one at a time."""

    def __init__(self, pipe: Union[Filter[Any,Any],Sink[Any]]):
        """Instantiate a Foreach pipe.

        Args:
            pipe: The pipe that we wish to pass a sequence of items one at a time.
        """
        self._pipe = pipe

    def filter(self, items: Iterable[Any]) -> Iterable[Any]:
        """Filter the items using the inner pipe. This method only works if the inner pipe is a Filter."""
        for item in items:
            yield self._pipe.filter(item)

    def write(self, items: Iterable[Any]):
        """Write the items using the inner pipe. This method only works if the inner pipe is a Sink."""
        for item in items:
            self._pipe.write(item)

    @property
    def params(self) -> Dict[str,Any]:
        return self._pipe.params

    def __str__(self):
        return str(self._pipe)

class QueueIO(Source[Iterable[Any]], Sink[Any]):
    def __init__(self, queue:Queue=None, block:bool=True, poison:Any=None) -> None:
        self._queue  = queue or Queue()
        self._sink   = QueueSink(queue)
        self._source = QueueSource(queue, block, poison)

    def write(self, item: Any) -> None:
        self._sink.write(item)

    def read(self) -> Iterable[Any]:
        return self._source.read()



from abc import abstractmethod, ABC
from typing import Any, Union, Iterable, Sequence, Mapping, overload

from coba.primitives import Context, Action, Actions
from coba.primitives import Reward, SequenceReward, Feedback, SequenceFeedback
from coba.pipes import Source, SourceFilters, Filter
from coba.exceptions import CobaException

class Interaction(dict):
    """An individual interaction that occurs in an Environment."""
    __slots__=()
    keywords = {}

    @property
    def extra(self) -> Mapping[str,Any]:
        return { k:self[k] for k in self.keys()-self.keywords}

class SimulatedInteraction(Interaction):
    """Simulated data that provides labels for every possible action."""
    __slots__=()
    keywords = {'type', 'context', 'actions', 'rewards'}
    
    def __init__(self,
        context : Context,
        actions : Actions,
        rewards : Union[Reward, Sequence[float]],
        **kwargs) -> None:
        """Instantiate SimulatedInteraction.

        Args
            context : Features describing the interaction's context.
            actions : Features describing available actions during the interaction.
            rewards : The reward for each action in the interaction.
            kwargs : Any additional information.
        """

        self['context'] = context
        self['actions'] = actions
        if isinstance(rewards,(list,tuple)):
            if len(rewards) != len(actions): 
                raise CobaException("The given actions and rewards did not line up.")
            self['rewards'] = SequenceReward(rewards)
        else:
            self['rewards'] = rewards

        if kwargs: self.update(kwargs)

class GroundedInteraction(Interaction):
    """Logged data providing a label for one action."""
    __slots__=()
    keywords = {'type', 'context', 'actions', 'rewards', 'feedbacks'}

    def __init__(self,
        context: Context,
        actions: Actions,
        rewards: Union[Reward, Sequence[float]],
        feedbacks: Union[Feedback, Sequence[Any]],
        **kwargs) -> None:
        """Instantiate GroundedInteraction.

        Args
            context: Features describing the interaction's context.
            actions: Features describing available actions during the interaction.
            rewards: The reward for each action in the interaction.
            feedbacks: The feedback for each action in the interaction.
            **kwargs: Additional information that should be recorded in the interactions table of an experiment result.
        """

        self['context'] = context
        self['actions'] = actions
        self['rewards'] = SequenceReward(rewards) if isinstance(rewards,(list,tuple)) else rewards
        self['feedbacks'] = SequenceFeedback(feedbacks) if isinstance(feedbacks,(list,tuple)) else feedbacks

        if kwargs: self.update(kwargs)

class LoggedInteraction(Interaction):
    """Logged data that describes an interaction where the choice was already made."""
    __slots__ = ()
    keywords = {'type', 'context', 'action', 'reward', 'probability', 'actions', 'rewards'}

    @overload
    def __init__(self,
        context: Context,
        action: Action,
        reward: float,
        *,
        probability: float=None,
        actions: Actions=None,
        rewards: Union[Reward, Sequence[float]] = None,
        **kwargs) -> None:
        ...

    def __init__(self,
        context: Context,
        action: Action,
        reward: float,
        **kwargs) -> None:
        """Instantiate LoggedInteraction.

        Args
            context: Features describing the logged context.
            action: Features describing the action taken by the logging policy.
            reward: The reward that was revealed when the logged action was taken.
            probability: The probability that the logged action was taken. That is P(action|context,actions,logging policy).
            actions: All actions that were availble to be taken when the logged action was taken. Necessary for OPE.
            rewards: The rewards to use for off policy evaluation. These rewards will not be shown to any learners. They will
                only be recorded in experimental results. If probability and actions is provided and rewards is None then 
                rewards will be initialized using the IPS estimator.
            **kwargs : Any additional information.
        """

        if 'actions' in kwargs and 'probability' in kwargs and 'rewards' not in kwargs:
            probability       = kwargs['probability']
            actions           = kwargs['actions']
            kwargs['rewards'] = [int(a==action)*reward/probability for a in actions]

        self['context'] = context
        self['action']  = action
        self['reward']  = reward

        if 'rewards' in kwargs and isinstance(kwargs['rewards'],(list,tuple)):
            kwargs['rewards'] = SequenceReward(kwargs['rewards'])

        if kwargs: self.update(kwargs)

class EnvironmentFilter(Filter[Iterable[Interaction],Iterable[Interaction]], ABC):
    """A filter that can be applied to an Environment."""

    @abstractmethod
    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
        """Apply a filter to an Environment's interactions."""
        ...

class Environment(Source[Iterable[Interaction]], ABC):
    """An Environment that produces Contextual Bandit data"""

    @property
    def params(self) -> Mapping[str,Any]: # pragma: no cover
        """Paramaters describing the simulation.

        Remarks:
            These will become columns in the environments table of experiment results.
        """
        return {}

    @abstractmethod
    def read(self) -> Iterable[Interaction]:
        """The sequence of interactions in the simulation.

        Remarks:
            This function should always be "re-iterable".
        """
        ...

    def __str__(self) -> str:
        return str(self.params) if self.params else self.__class__.__name__

class SimulatedEnvironment(Environment):
    """An environment made from SimulatedInteractions."""

    @abstractmethod
    def read(self) -> Iterable[SimulatedInteraction]:
        """The sequence of interactions in the environment.

        Remarks:
            This function should always be "re-iterable".
        """
        ...

class SafeEnvironment(Environment):
    """A wrapper for environment-likes that guarantees interface consistency."""

    def __init__(self, environment: Environment) -> None:
        """Instantiate a SafeEnvironment.

        Args:
            environment: The environment we wish to make sure has the expected interface
        """

        self._environment = environment if not isinstance(environment, SafeEnvironment) else environment._environment

    @property
    def params(self) -> Mapping[str, Any]:
        try:
            params = self._environment.params
        except AttributeError:
            params = {}

        if "type" not in params:

            if isinstance(self._environment, SourceFilters):
                params["type"] = self._environment._source.__class__.__name__
            else:
                params["type"] = self._environment.__class__.__name__

        return params

    def read(self) -> Iterable[Interaction]:
        return self._environment.read()

    def __str__(self) -> str:
        params = dict(self.params)
        tipe   = params.pop("type")

        if len(params) > 0:
            return f"{tipe}({','.join(f'{k}={v}' for k,v in params.items())})"
        else:
            return tipe



import gzip

from queue import Queue
from collections.abc import Iterator
from typing import Callable, Any, List

from coba.pipes.primitives import Sink

class NullSink(Sink[Any]):
    """A sink which does nothing with written items."""
    def write(self, item: Any) -> None:
        pass

class ConsoleSink(Sink[Any]):
    """A sink which prints written items to console."""

    def write(self, item: Any) -> None:
        print(item)

class DiskSink(Sink[str]):
    """A sink which writes to a file on disk.

    This sink supports writing in either plain text or as a gz compressed file.
    In order to make this distinction gzip files must end with a gz extension.
    """

    def __init__(self, filename:str, mode:str='a+') -> None:
        """Instantiate a DiskSink.

        Args:
            filename: The path to the file to write.
            mode: The mode with which the file should be written.
        """

        #If you are using the gzip functionality of disk sink
        #then you should note that this implementation isn't optimal
        #in terms of compression since it compresses one line at a time.
        #see https://stackoverflow.com/a/18109797/1066291 for more info.

        self._filename = filename
        self._count    = 0
        self._file     = None
        self._mode     = mode

    def __enter__(self) -> 'DiskSink':
        self._count += 1

        if self._file is None:
            if ".gz" in self._filename:
                self._file = gzip.open(self._filename, f"{self._mode}b", compresslevel=6)
            else:
                self._file = open(self._filename, f"{self._mode}b")

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._count -= 1
        if self._count == 0 and self._file is not None:
            self._file.close()
            self._file = None

    def write(self, item: str) -> None:
        with self:
            self._file.write((item + '\n').encode('utf-8'))
            self._file.flush()

class ListSink(Sink[Any]):
    """A sink which appends written items to a list."""

    def __init__(self, items: List[Any] = None) -> None:
        """Instantiate a ListSink.

        Args:
            items: The list we wish to write to.
        """
        self.items = items if items is not None else []

    def write(self, item) -> None:
        self.items.append(list(item) if isinstance(item, Iterator) else item)

class QueueSink(Sink[Any]):
    """A sink which puts written items into a Queue."""

    def __init__(self, queue:Queue=None) -> None:
        """Instantiate a QueueSink.

        Args:
            queue: The queue to put written items into.
        """
        self._queue  = queue or Queue()

    def write(self, item: Any) -> None:
        try:
            self._queue.put(item)
        except (EOFError,BrokenPipeError):
            pass

class LambdaSink(Sink[Any]):
    """A sink which passes written items to a callable function."""

    def __init__(self, write: Callable[[Any],None]):
        """Instantiate a LambdaSink.

        Args:
            write: A callable function written items will be passed to.
        """
        self._write = write

    def write(self, item: Any) -> None:
        return self._write(item)



import re
import csv

from operator import methodcaller
from collections import deque
from itertools import islice, chain, count, takewhile
from typing import Iterable, Sequence, Union, Any, Pattern, Tuple, Mapping
from typing import MutableSequence, MutableMapping, Callable, Iterator

from coba.exceptions import CobaException
from coba.encodings import Encoder, CategoricalEncoder
from coba.utilities import peek_first
from coba.primitives.rows import Dense, Sparse

from coba.pipes.rows import HeadRows, LazyDense, LazySparse
from coba.pipes.primitives import Filter

class CsvReader(Filter[Iterable[str], Iterable[MutableSequence]]):
    """A filter capable of parsing CSV formatted data."""

    def __init__(self, has_header: bool=False, **dialect):
        """Instantiate a CsvReader.

        Args:
            has_header: Indicates if the CSV data has a header row.
            **dialect: This has the same values as Python's csv.reader dialect.
        """
        self._dialect    = dialect
        self._has_header = has_header

    def filter(self, items: Iterable[str]) -> Iterable[Dense]:

        lines = iter(csv.reader(iter(filter(None,(i.strip() for i in items))), **self._dialect))
        first = next(lines)

        if self._has_header:
            return HeadRows(first).filter(lines)
        else:
            return chain([first],lines)

class ArffAttrReader(Filter[Iterable[str], Iterable[Tuple[str,Callable]]]):

    class CategoricalDict(dict):
        def __missing__(self, key):
            raise CobaException(f"We were unable to find '{key}' in {sorted(self.keys())}.")

    def __init__(self, is_dense:bool) -> None:
        self._is_dense = is_dense
        self._r_space  = re.compile("(\s+)")
        self._r_comma  = re.compile("(,)")

    def filter(self, lines: Iterable[str]) -> Iterable[Tuple[str,Callable]]:

        headers = set()

        for line in lines:
            if line[0:10].lower() == "@attribute":
                header, encoding = tuple(self._split(line[11:], self._r_space, n=2))

                if header in headers:
                    raise CobaException("Two columns in the ARFF file had identical header values.")
                else:
                    headers.add(header)

                yield header, self._encoder(encoding)

    def _split(self, line: str, pattern: Pattern[str], n=None):

        items  = iter(pattern.split(line))
        count  = 0
        quotes = '"'+"'"

        try:
            while True:

                item = next(items).lstrip()
                if not item or pattern.match(item): continue

                count += 1
                if count == n:
                    items = chain([item],items)
                    break

                if item[0] in quotes:
                    q  = item[0]
                    while item.rstrip()[-1] != q or item.rstrip()[-2]=="\\":
                        item += next(items)

                    item = item.strip().rstrip()[1:-1].replace("\\",'')
                else:
                    item = item.strip()

                yield item

            yield "".join(items).strip()
        except StopIteration:
            pass

    def _encoder(self, encoding: str) -> Encoder:
        numeric_types = ('numeric', 'integer', 'real')
        string_types  = ("string", "date", "relational")

        if encoding.lower() in numeric_types:
            return float

        elif encoding.lower().startswith(string_types):
            return lambda x: None if x == "?" else x

        elif encoding.startswith('{'):
            categories = list(self._split(encoding[1:-1], self._r_comma))

            if not self._is_dense:
                #there is a bug in ARFF where the first class value in an ARFF class can will dropped from the
                #actual data because it is encoded as 0. Therefore, our ARFF reader automatically adds a 0 value
                #to all sparse categorical one-hot encoders to protect against this.
                categories = ["0"] + categories

            return ArffAttrReader.CategoricalDict(CategoricalEncoder(categories)._categoricals).__getitem__

        else:
            raise CobaException(f"An unrecognized encoding was found in the arff attributes: {encoding}.")

class ArffDataReader(Filter[Iterable[str], Iterable[Union[Dense,Sparse]]]):

    _trans = str.maketrans('','',' \t\n\r\v\f')

    def __init__(self, is_dense:bool) -> None:
        self._is_dense = is_dense

    def filter(self, lines: Iterable[str]) -> Iterable[Tuple[str,bool]]:
        return (self._dense if self._is_dense else self._sparse)(lines)

    def _dense(self, lines: Iterable[str]) -> Iterable[Tuple[str,bool]]:
        for line in lines:
            if line[0] == "%": continue
            if "?" not in line:
                missing = False
            elif line[:2] == "?,":
                missing = True
            elif line[-2:] == ",?":
                missing = True
            else:
                compact = line.translate(self._trans)
                missing = compact[:2] == '?,' or ',?,' in compact or compact[-2:] == ',?'

            yield line,missing

    def _sparse(self, lines: Iterable[str]) -> Iterable[Tuple[str,bool]]:

        for line in lines:
            if line[0] == "%": continue
            missing = " ?," in line or line[-3:] == " ?}"
            yield line,missing

class ArffLineReader(Filter[str, Sequence[str]]):
    def __init__(self, is_dense:bool, n_columns:int, ):
        self._is_dense = is_dense
        self._n_columns = n_columns

        #only used when parsing dense lines
        self._fallback_delim = None
        self._quotes         = '"'+"'"
        self._dialect        = dict(skipinitialspace=True,escapechar="\\",doublequote=False)
        self._quotechar      = None

        if self._is_dense:
            self._set_filter(self._dense)
        else:
            self._set_filter(self._sparse)

    def _set_filter(self,method)->None:
        self.filter = method

    def filter(self,line:str) -> Union[Sequence[str],Mapping[str,str]]:
        #this is defined in __init__ for performance purposes 
        pass #pragma: no cover

    def _dense(self, line:str) -> Sequence[str]:

        self._set_filter(self._dense_simple)

        dialect = self._dialect

        double_quote_in_line = '"' in line
        single_quote_in_line = "'" in line

        if single_quote_in_line and double_quote_in_line:
            self._set_filter(self._dense_advanced)
        elif double_quote_in_line:
            dialect['quotechar'] = '"'
            self._quotechar      = '"'
        elif single_quote_in_line:
            dialect['quotechar'] = "'"
            self._quotechar      = "'"

        if len(next(csv.reader([line], **dialect, delimiter=","))) == self._n_columns:
            dialect['delimiter'] = ','
        elif len(next(csv.reader([line], **dialect, delimiter='\t'))) == self._n_columns:
            dialect['delimiter'] = '\t'
        else:
            self._set_filter(self._dense_advanced)

        return self.filter(line)

    def _sparse(self, line:str) -> Mapping[int,str]:
        keys_and_vals = re.split('\s*,\s*|\s+', line.strip("} {"))

        if keys_and_vals != ['']:
            keys = list(map(int,keys_and_vals[0::2]))
            vals = keys_and_vals[1::2]
        else:
            keys = []
            vals = []

        parsed = dict(zip(keys,vals))
        if parsed and (min(parsed.keys()) < 0 or self._n_columns <= max(parsed.keys())):
                raise CobaException(f"We were unable to parse a line in a way that matched the expected attributes.")
        return parsed

    def _dense_simple(self,line:str)-> Sequence[str]:
        dialect = self._dialect
        quotechar = self._quotechar

        if '"' in line:
            if quotechar == '"':
                pass
            elif quotechar is None:
                self._quotechar = '"'
                dialect['quotechar'] = '"'
            else:
                self._set_filter(self._dense_advanced)
                return self.filter(line)

        if "'" in line:
            if quotechar == "'":
                pass
            elif quotechar is None:
                self._quotechar = "'"
                dialect['quotechar'] = "'"
            else:
                self._set_filter(self._dense_advanced)
                return self.filter(line)

        parsed = next(csv.reader([line], **dialect))
        if len(parsed) != self._n_columns:
            raise CobaException(f"We were unable to parse a line in a way that matched the expected attributes.")
        return parsed

    def _dense_advanced(self,line:str) -> Sequence[str]:
        #the file does not appear to follow a readable csv.reader dialect
        #we fall back now to a slightly slower, but more flexible, parser

        if self._fallback_delim is None:
            #this isn't airtight but we can only infer so much.
            self._fallback_delim = ',' if len(line.split(',')) > len(line.split('\t')) else "\t"

        d_line = deque(line.split(self._fallback_delim))
        parsed = []

        while d_line:
            item = d_line.popleft().lstrip()

            if item[0] in self._quotes:
                possible_quotechar = item[0]
                while item.rstrip()[-1] != possible_quotechar or item.rstrip()[-2] == "\\":
                    item += "," + d_line.popleft()
                item = item.strip()[1:-1]

            parsed.append(item.replace('\\',''))

        if len(parsed) != self._n_columns:
            raise CobaException(f"We were unable to parse a line in a way that matched the expected attributes.")
        return parsed

class ArffReader(Filter[Iterable[str], Iterable[Union[Dense,Sparse]]]):
    
    _strip = methodcaller("strip")

    def __init__(self):
        """Instantiate an ArffReader."""

    def filter(self, lines: Iterable[str]) -> Iterable[Union[Dense,Sparse]]:

        lines = iter(filter(None,map(self._strip, lines)))
        attrs = [ l for l in takewhile(lambda l: l.lower()!="@data", lines) if l[:5].lower() == "@attr" ]
        data  = lines

        first_data,data = peek_first(data)
        while data and first_data[0]=="%":
            first_data,data = peek_first(islice(data,1,None))

        if not data: return []

        is_dense = not first_data.startswith("{") or not first_data.endswith("}")

        attr_reader = ArffAttrReader(is_dense)
        data_reader = ArffDataReader(is_dense)
        line_reader = ArffLineReader(is_dense,len(attrs))

        headers,encoders = zip(*attr_reader.filter(attrs))

        if is_dense:
            hdr_map = dict(zip(headers, count()))
            for line,missing in data_reader.filter(data):
                yield LazyDense(lambda line=line:line_reader.filter(line), encoders,hdr_map, missing)

        else:
            encs = dict(enumerate(encoders))
            fwd  = dict(zip(headers,count()))
            inv  = dict(zip(count(),headers))
            nsp  = set()
            for k,v in encs.items():
                try:
                    if v('0')!=0: nsp.add(k)
                except: pass #pragma: no cover

            for line,missing in data_reader.filter(data):
                yield LazySparse(lambda line=line:line_reader.filter(line), encs, nsp, fwd, inv, missing)

class LibsvmReader(Filter[Iterable[str], Iterable[Tuple[MutableMapping,Any]]]):
    """A filter capable of parsing Libsvm formatted data.

    For a complete description of the libsvm format see `here`__ and `here`__.

    __ https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets/
    __ https://github.com/cjlin1/libsvm
    """

    def filter(self, lines: Iterable[str]) -> Iterable[Tuple[MutableMapping, Any]]:

        for line in filter(None,lines):

            items  = line.strip().split(' ')

            no_label_line = items[0] == '' or ":" in items[0]

            if not no_label_line:
                labels = items[0].split(',')
                row    = { int(k):float(v) for i in items[1:] for k,v in [i.split(":")] }
                yield (row, labels)

class ManikReader(Filter[Iterable[str], Iterable[Tuple[MutableMapping,Any]]]):
    """A filter capable of parsing Manik formatted data.

    For a complete description of the manik format see `here`__ and `here`__.

    __ http://manikvarma.org/downloads/XC/XMLRepository.html
    __ https://drive.google.com/file/d/1u7YibXAC_Wz1RDehN1KjB5vu21zUnapV/view
    """

    def filter(self, lines: Iterable[str]) -> Iterable[Tuple[MutableMapping, Any]]:
        # we skip first line because it just has metadata
        return LibsvmReader().filter(islice(lines,1,None))



from coba.backports.metadata import version, entry_points, PackageNotFoundError
from coba.backports.typing import Literal

__all__ = [
    'version',
    'entry_points',
    'PackageNotFoundError',
    'Literal',
]